# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 12:09:38 2016

@author: akidribm
"""
# Imports necessary sort types and functions from the library
from sortal import vector2D,vector2Dfloat, planeSegment2D, primitiveSort, attributeSort, disjunctiveSort, weight, point2D, coForm, label, affineTransform, transformSet
from rule5_RotatingPlane_Points_Weight_Label import rule5a,rule5b

# Imports math functions
from decimal import Decimal
import math

# Imports API functions and variables
from api_create import create
from transform import transform
from sort_types import sort_types

# Imports GUI library
from pylab import *
import matplotlib
from sympy import pi,cos,sin
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import  matplotlib.pylab as plt
import matplotlib.pyplot as pltplot
from matplotlib.collections import LineCollection
matplotlib.use ("gTkAgg")
from matplotlib.collections import PatchCollection
import tkinter as tk
from tkinter import ttk, Tk, Label, Button, Frame
from tkinter import *

plt.ion()

"""
=============
FORM CREATION
=============
'Create' methods generate and return forms based on input. As part of the API, they are imported by 
importing create from the api_create file, i.e. from api_create import create.

All sort types are initialized and stored in the file sort_types.py, inside the dictionary sort_types.
This is imported via 'from sort_types import sort_types'

Vectors are created by calling either vector2D or vector3D, as follows:
v1 = vector2D(x,y) where (x,y) are coordinates
v2 = vector3D(x,y,z) where (x,y,z) are coordinates

=======
METHODS
=======              
Create.planeForm where a form bounded by the elements in the bound argument with the sort_type (planeSort) is created. 
                 Attributes can be added later on if the planeSort2D defined in sort_types is based on attributeSort,
                 e.g. planeSort2D = attributeSort('planeAttrib2D', planeS2D, weightSort) 
                 OR planeSort2D = attributeSort('planeAttrib2D', planeS2D, labelSort) 
                 
                 This method will raise an error if the sort type passed is not based on planeSegment2D.
                
                 bound = [v1, v2, v3, v4]
                 formP = create.planeForm(sort_types['planeSort2D'], bound)
                 In this case planeSort is an attribute sort with a weight attribute (as defined inside the file sort_types)
                 
                 To add an attribute like weight to the first/nth plane inside formP:
                     
                 Create the weight attribute first: weight = weight(sort_types['weight'], 255)
                 
                 The plane/s inside formP are housed inside formP.elements
                 EXAMPLE: To add an attribute to the first plane inside form: formP.elements[0].addAttribute(weight)  

Create.pointForm where a form of points is created from the inputs sort_type, and the (x,y) or (x,y,z) coordinates of an 
                 initial point which is inserted in the form. New points can be created and inserted later 
                 on into the form. This method will raise an error if the sort type passed is not based on point2D or point3D.

                 formPt = create.pointForm (sort_types['pointSort2D'], 1, 1) OR formPt = create.pointForm (sort_types['pointSort3D'], 1, 1, 1)
                 
                 To insert another point,
                 point = point2D(sort_types['pointSort2D'], 1, 0) OR point = point3D(sort_types['pointSort3D'], 1, 0, 1)
                 formPt.insert(point)

Create.addAttributeAll takes a form and an attribute as inputs. It adds the attribute (weight/label) to 
                       all points inside the form.
                       
                       This method will raise an error if the sort type of the form is not based on attributeSort.
                       Note: A check for the attribute sort type will be added later on.
                    
                       1. Create label attribute
                       label = label(sort_types['label'], 'black')
                    
                       2. Initially add one point at (1,1) to the form --> form has 1 point: (1,1)
                       formPt = create.pointForm (sort_types['pointSort2D'], 1, 1) 
                       
                       Transform point inside point form by angle a
                       formPt1 = transform(formPt1, a) 

                       3. Add the other point at (1,1) to the form --> form has 2 points: (sqrt(3)/3,1) and (1,1)
                       point2 = point2D(sort_types['pointSort2D'], 1, 1)   
                       formPt.insert(point2) 
                    
                       4. Add black label to both points
                       formPt = create.addAttributeAll(formPt, label) 

Create.metaForm takes a list of forms as input and inserts them into a disjunctive Sort metaform. This method will raise an
                error if the sort type passed is not based on disjunctiveSort.

                forms = [formP1, formPt1]
                formM1 = create.metaForm(sort_types['disjunctive2D'], forms)
"""
if __name__ == "__main__":
    
    LARGE_FONT= ("Verdana", 16)
    NORM_FONT= ("Verdana", 12)
    
    a = 15 # Angle to be used for transform function 

    """
    ========
    EXAMPLES
    ========
    """
    create = create() # Initializes 'create' methods
    
    weight1 = weight(sort_types['weight'], 255) # Initializes weight (black)
    
    v1 = vector2D(-1, -1)
    v2 = vector2D(-1, 1)
    v3 = vector2D(1, 1)
    v4 = vector2D(1, -1)
    
    bound = [v1,v2,v3,v4] 
    
    # PLANE WITH WEIGHT ATTRIBUTE
    # Creates plane from sort_types['planeSort2D'], boundary vectors and weight1
    formP1 = create.planeForm(sort_types['planeSort2D'], bound)
    formP1.elements[0].addAttribute(weight1)
    
    # POINT FORM WITH LABEL ATTRIBUTE
    # Initializes label (text: 'white')
    label1 = label(sort_types['label'], 'white') 
    # Initializes point with (x,y) = (1,1)
    point1 = point2D(sort_types['pointSort2D'], 1, 1)   
    
    # Creates form with an initial point (x,y) = (1,1) inside
    formPt1 = create.pointForm (sort_types['pointSort2D'], 1, 1) 
    # Transforms point by angle a, which moves point to a different location 
    formPt1 = transform(formPt1, a) 
    # Inserts another point with (x,y) = (1,1), so now there are two points: the transformed point and the new point
    formPt1.insert(point1)
    # Adds white label to both points inside the form (the initial transformed point and the inserted point)
    formPt1 = create.addAttributeAll(formPt1, label1) 
    
    # DISJUNCTIVE FORM WITH PLANE AND POINTS
    # List of forms to be put inside metaform
    forms = [formP1, formPt1] 
    # Creates metaform of sort type 'disjunctive2D' and with forms from the list 'forms' inside
    formM1 = create.metaForm(sort_types['disjunctive2D'], forms) 
    
    # Initializes rules 5a and 5b with angle a as input
    r1 = rule5a(a)
    r2 = rule5b(a)
            
    class Application(Frame):
        
        print ("<<<< GENERATING APPLICATION WINDOW >>>>")
        fig_main = Figure(figsize = (8, 8), dpi = 150) # Figure contains 'Current Drawing'
        fig_rules = Figure(figsize = (6.5, 6.5), dpi = 100) # Figure contains 'Rules'
        blackPlanes = [] # List contains all planes with weight '255' 
        whitePlanes = [] # List contains all planes with weight '0'
        count = 1 # Keeps track of which rule is to be applied
        
        def __init__(self, master = None):
            Frame.__init__(self, master)
            self.grid()
            self.master.title("Test Case 2: Shape Grammar Interpreter")
    
            # Generates tkinter window grid with 6 rows and 6 columns
            for r in range(6):
                self.master.rowconfigure(r, weight=1)    
            for c in range(6):
                self.master.columnconfigure(c, weight=1)

            # Holds 'Current Drawing' figure (fig_main)
            self.Frame1 = Frame(master, bg="gray")
            self.Frame1.grid(row = 0, column = 0, rowspan = 6, columnspan = 5, sticky = W+E+N+S) 
            self.Frame1_label = Label (master = self.Frame1, text = "Current Drawing, Iteration {}".format(self.count), background = "gray", font = LARGE_FONT, anchor = "center")
            self.Frame1_label.pack()
            self.Frame2_label = Label(master = self.Frame1, text = "Matches are highlighted in orange when the cursor hovers over the plot. \nClick directly on the match plane to apply the relevant rule.", background = 'orange', font = NORM_FONT, anchor = "center")
            self.Frame2_label.pack(fill = BOTH, expand = 1)
            
            # Holds 'Rules' figure (fig_rules)
            # LHS and RHS figures are side by side for each rule
            self.Frame3 = Frame(master, bg='gray')
            self.Frame3.grid(row = 0, column = 5, rowspan = 6, columnspan = 1, sticky = W+E+N+S)
            self.Frame3_label = Label (master = self.Frame3, text ="Rules", background = 'gray', font = LARGE_FONT, anchor = "center")
            self.Frame3_label.pack()
            
            self.menu() # Initializes menu
            self.set_figure() # Embeds matplotlib figures 'fig_main' and 'fig_rules' in tkinter window
            self.call_axes() # Initalizes subplot in 'fig_main' and adjusts plot settings
            self.initialization() # Plots initial form onto 'fig_main', 'a_main'
            self.connect() # Activates mouseover, mouseoff and mouseclick functions
            self.rules_viz() # Plots rules 1 and 2 in RHS of tkinter window
            
        """
        =====================
        API: RULE APPLICATION
        =====================
        
        Generate(count) takes the current self.count value as input and applies r1 (even count value) or r2 (odd count value). 
        This generates the matches as well as the new planes in the form and changes the form's contents.
        The r1.detect/r2.detect method takes a form as input and detects matches in the form. The r1.apply/r2.apply method 
        combines both detect and perform methods (transforming the match from LHS to RHS).
        """
        def generate(self, count):
            if count%2 == 0:
                r1.apply(formM1)  
                r2.detect(formM1) # For purposes of showing match in GUI
            else:
                r2.apply(formM1) 
            
        """
        ========================
        GRAPHICAL USER INTERFACE
        ========================
        """
        # Sets up and formats main figure plot inside self.fig_main
        def call_axes(self):
            self.a_main = self.fig_main.add_subplot(111)
            self.a_main.axes.get_xaxis().set_visible(False)
            self.a_main.axes.get_yaxis().set_visible(False) 
            self.a_main.margins(y=0.2)
            self.a_main.margins(x=0.2)
            self.a_main.axes.set_aspect('equal')
            self.a_main.set_xlim (-1.2,1.2)
            self.a_main.set_ylim (-1.2,1.2)
            
        # Embeds main figure (self.fig_main) and rules figure (self.fig_rules) in tkinter window
        def set_figure(self):
            self.canvas_main = FigureCanvasTkAgg(self.fig_main, self.Frame1)
            self.canvas_main.get_tk_widget().pack(fill = BOTH, expand = 1) 
            
            self.canvas_rules = FigureCanvasTkAgg(self.fig_rules, self.Frame3)
            self.canvas_rules.get_tk_widget().pack(fill = BOTH, expand = 1) 

        # Plots initial figure (black square with two points and two labels)
        def initialization (self):        
            self.rearrange_planes()
            pts = self.data_prep(self.rearranged_list[0])
            patch = plt.Polygon(pts, facecolor = 'black')
            self.a_main.add_patch(patch)
            self.over_patch = patch
            self.draw_point_label()
        
        # Plots rules 1 and 2 in Frame 3
        def rules_viz(self):
            ind = 1
            extra_label = ''
            for i in range (4):
                self.a_rules = self.fig_rules.add_subplot(2, 2, i+1)
                self.a_rules.axes.get_xaxis().set_visible(False)
                self.a_rules.axes.get_yaxis().set_visible(False) 
                self.a_rules.margins(y=0.2)
                self.a_rules.margins(x=0.2)
                self.a_rules.axes.set_aspect('equal')
                self.a_rules.set_xlim (-1.2,1.2)
                self.a_rules.set_ylim (-1.2,1.2)
                 
                if i <= 1:
                    ind = 1 
                    rule_point = r1
                else:
                    ind = 2
                    rule_point = r2
                
                if i%2 == 0:
                    extra_label = "LHS\n"
                    rule_side = rule_point.lhs
                else:
                    extra_label = "RHS\n"
                    rule_side =  rule_point.rhs

                self.a_rules.set_title(extra_label + 'Rule {}'.format(ind), fontsize = 11)
                
                # Plots planes in rules
                for k in rule_side.elements[1]:
                    pts = self.data_prep(k)
                    if k.attribute.individual.value == 255:
                         patch = plt.Polygon(pts, facecolor = 'black')
                    else:
                        patch = plt.Polygon(pts, facecolor = 'white')
                    self.a_rules.add_patch(patch)
                
                # Plots points in rules
                for j in rule_side.elements[0]:
                    if i == 1 or i == 2:
                        self.a_rules.plot(j.value.x, j.value.y, 'o', c = 'black', markersize = 5, markeredgecolor='orange', markeredgewidth=0.5) 
                    else:
                        self.a_rules.plot(j.value.x, j.value.y, 'wo', markersize = 5, markeredgecolor='orange', markeredgewidth=0.5) 
                    # Annotates points with labels
                    self.a_rules.annotate(j.attribute.elements[0].value, xy = (j.value.x, j.value.y), xycoords='data', xytext=(-2.7, 2.7), size = 6.5, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'orange')) 
                
        # Draws plane match when mouse cursor is hovering over axis
        # Called in self.on_motion()
        def viz_match (self, matchF, face_color, edge_color = 'orange'):  
            pts = self.data_prep(matchF)
            patch = plt.Polygon(pts, facecolor = face_color, edgecolor = edge_color, linewidth = 1.5)
            self.a_main.add_patch(patch)
            self.fig_main.canvas.draw_idle()
            
        # Plots result and changes frame label
        def viz_result (self):
            patchList = []
            self.rearrange_planes()
            
            if self.count%2 == 0: # Determines while rule has been applied
                ind = 1
            else: 
                ind = 2
                
            self.Frame1_label.configure(text = "Current Drawing, Iteration {}".format(self.count)+", applied Rule {}".format(ind))
            
            for c, i in enumerate(self.rearranged_list): # Plots form planes after rule has been applied
                pts = self.data_prep(i)
                patch = plt.Polygon(pts, facecolor = self.color_list[c])
                self.a_main.add_patch(patch)
                patchList.append(patch)
                
            # Determines which plane is the match in the next iteration
            if self.count%2 == 0: 
                self.over_patch = patchList[-1]
            else: 
                self.over_patch = patchList[len(self.blackPlanes) - 1]
            
            self.draw_point_label() # Plots points with their accompanying labels
            self.fig_main.canvas.draw_idle() # Redraws plot
        
        # Draws point/s and accompanying label/s onto plane/s
        def draw_point_label(self):
            for i in formM1.elements[0].elements:
                # Adjusts label size and point size depending on iteration, as squares becomes smaller
                if self.count >= 6:
                    size_label = 4.5
                    marker_size = 3
                else:
                    size_label = 6
                    marker_size = 5.5
                    
                # Changes color of point based on count number (and thus, rule being applied)    
                if self.count%2 == 0:
                    self.a_main.plot(i.value.x, i.value.y, 'o', c = 'black', markersize = marker_size, markeredgecolor='orange', markeredgewidth=0.5) 
                else:
                    self.a_main.plot(i.value.x, i.value.y, 'wo', markersize = marker_size, markeredgecolor='orange', markeredgewidth=0.5)  
                # Annotates points with label
                self.a_main.annotate(i.attribute.elements[0].value, xy = (i.value.x, i.value.y), xycoords='data', xytext=(-2.7, 2.7), size = size_label, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'orange')) 
        
        # Separates black and white planes from each other and recombines them as one list for easier calling
        def rearrange_planes(self):
            self.blackPlanes = []
            self.whitePlanes = []
            color_list1 = []
            color_list2 = []

            for count, plane in enumerate(list(reversed(formM1.elements[1].elements))):
                if  plane.attribute.individual.value == 255:
                    self.blackPlanes.append(plane)
                    color_list1.append('black')
                elif  plane.attribute.individual.value == 0:
                    self.whitePlanes.append(plane) 
                    color_list2.append('white')
                
            self.rearranged_list = self.blackPlanes + self.whitePlanes
            self.color_list = color_list1 + color_list2
         
        # Prepares plane vertices for plotting as patches; appends each vector coordinate pair to vertex_list  
        def data_prep(self, vertex_source):
            vertex_list = []
            for j in vertex_source.boundaries[0].polygonal:
                vertex_list.append([j.x, j.y])
            return vertex_list

        # Connects mouse event functions to the 'Current Drawing' figure (fig_main)
        def connect(self):
            self.cidpress = self.fig_main.canvas.mpl_connect('button_press_event', self.on_press)
            self.cidmotion_on = self.fig_main.canvas.mpl_connect('axes_enter_event', self.on_motion)
            self.cidmotion_off = self.fig_main.canvas.mpl_connect('axes_leave_event', self.off_motion)
         
        # Calls functions to draw result of rule application 
        def on_press(self, event):
            if self.over_patch.contains(event)[0]:
                self.count += 1
                self.generate(self.count) # Applies rule based on current count
                self.fig_main.clf() # Clears figure
                self.call_axes() # Redraws axis
                self.viz_result() # Draws result 
            else:
                self.error_box()

        # Identifies which subplot the mouse is hovering over, changes edgecolor of plane match to orange
        def on_motion(self, event):
            if event.inaxes == self.a_main:
                if self.count%2 == 0:
                    self.viz_match(r2.detection.matches()[0].elements[1].elements[0], face_color = 'white') # Shows rule 2 match
                else: 
                    r1.detect(formM1) # Detects rule 1 match in formM1
                    self.viz_match(r1.detection.matches()[0].elements[1].elements[0], face_color = 'black') # Shows rule 1 match
                
        # Removes orange color from edges of plane match when  mouse cursor is not hovering over axis
        def off_motion(self, event):
            if event.inaxes == self.a_main:
                if self.count%2 == 0:
                    self.viz_match(r2.detection.matches()[0].elements[1].elements[0], face_color = 'white', edge_color = 'white')
                else: 
                    self.viz_match(r1.detection.matches()[0].elements[1].elements[0], face_color = 'black', edge_color = 'black')
                
        # Generates error box if mouse click is not within area of match
        def error_box(self):
            error_box = Tk()
            error_box.title("Input Error")
            Label(error_box, text = "Please click directly on the match.", padx = 5, pady = 15).pack()
            Button(error_box, text= "OK", padx = 5, pady = 5, command= error_box.destroy).pack()  
            error_box.update()

        # Menu Commands: Save - saves PDF of 'Current Drawing' figure, Exit - closes window
        def menu(self):
            menubar = Menu (root)  # Saves PDF of 'Current Drawing' only
            menubar.add_command(label="Save Current Drawing", command=self.save) # Closes window
            menubar.add_command(label="Exit", command=root.destroy)
            root.config(menu=menubar)
            
        # Saves file of 'Current Drawing'
        # File type may be changed by change the .pdf extension to e.g.: .jpg, .png 
        def save(self):
            self.fig_main.savefig('currentdrawing.pdf', bbox_inches='tight')
            
    root = Tk()
    # Adjusts window size; width x height + x + y (x,y coordinates for where window will be placed on screen)
    root.geometry("1300x700+30+20")
    app = Application(master=root)
    # Calls and opens GUI
    app.mainloop()
    
