# -*- coding: utf-8 -*-
"""
Created on Mon Aug 22 15:24:49 2016

@author: akibdpt
"""

from .multiplyForm import multiplyForm
from copy import deepcopy

"""
 * INTERVAL is a behavioral category for Sorts (see 'Sort.java').
 * An interval sort is a multiply-associated sort for which
 * ...
 * Therefore, the IntervalForm class extends on the MultiplyForm
 * class (see 'MultiplyForm.java') and implements the algebra methods
 * from the Form interface (see 'Form.java').
 *
 * ASSUMPTIONS: the following algebraic property is assumed on forms
 *   a + b = b <=> a < b
 * 
"""
class interval(multiplyForm):
    
    def __init__(self, sort):
        super().__init__(sort)
        self.type = 'form-interval'
    
    """
    print out the form for easier assessment
    structure:
    |sort name: n elements
     _ individual 1
     _ individual 2
         ...
     _ individual n
      |attribute1 sort name: m elements
       _ attribute1 1
      |attribute2 sort name: p elements
       _ attribute1 2
    """
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' _ ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return
        
    """
    inserting an element into the form
    """
    def insert(self, ind):
        if not self.sort == ind.sort:
            raise ValueError('not of the same sort')
            return
        thisInd = self.first()
        index = 0
        while thisInd is not None and thisInd.compareInterval(ind) < 1:
            if thisInd.compareInterval(ind) == 0:
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            else:
                thisInd = self.next(thisInd)
                index = index+1
        self.add(index, ind)
        return
        
        
    def oneStepBack(self, current):
        prevInd = self.previous(current)
        if prevInd is not None and current.touches(prevInd) and prevInd.attribute == current.attribute:
            prevInd.combine(current)
            self.remove(current)
        return prevInd
        
    def oneStepForward(self, current):
        nextInd = self.next(current)
        if nextInd is not None and current.touches(nextInd) and nextInd.attribute == current.attribute:
            nextInd.combine(current)
            self.remove(current)
        return nextInd
        
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    def maximalize(self):
        if self.maximal:
            return
    
        self.maximal = True

        if len(self.elements) == 0:
            return

        ind = self.first()
        nextInd = self.next(ind)
        disorganized = []
        while nextInd is not None:
            if nextInd.isNil():
                self.remove(nextInd)
                ind = nextInd
                nextInd = self.next(nextInd)
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.next(nextInd)
            elif ind.compareValue(nextInd) == 0:
                if ind.compareInterval(nextInd) == 0:
                    if ind.hasAttribute():
                        ind.attribute.sum(nextInd.attribute)                
                    toBeRemoved = nextInd                
                    nextInd = self.next(nextInd)
                    self.remove(toBeRemoved)
                elif ind.touches(nextInd) or ind.disjoint(nextInd):
                    if ind.compareInterval(nextInd) == 1:
                        toBeRemoved = nextInd                
                        nextInd = self.next(nextInd)
                        disorganized.append(toBeRemoved)
                        self.remove(toBeRemoved)
                    else:
                        ind = nextInd
                        nextInd = self.next(nextInd)  
                    
                elif ind.contains(nextInd):
                    if not ind.hasAttribute() or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        nextInd = self.next(ind)
                    else:
                        temp = ind.complement(nextInd)
                        if ind.hasAttribute():
                            nextInd.attribute.sum(ind.attribute)
                        toBeRemoved = ind             
                        
                        if temp[0].compareValue(nextInd) == -1:
                            ind = temp[0]
                        else: 
                            ind = nextInd
                            
                        self.remove(toBeRemoved)
                        
                        for i in temp:
                            self.insert(i)
                        nextInd = self.next(ind)
                        
                elif nextInd.contains(ind):
                    if not ind.hasAttribute() or ind.attribute == nextInd.attribute:
                        self.remove(ind)
                        ind = nextInd
                        nextInd = self.next(nextInd)
                    else:
                        temp = nextInd.complement(ind)
                        if nextInd.hasAttribute():
                            ind.attribute.sum(nextInd.attribute)
                        if temp[0].compareValue(ind) == -1:
                            ind = temp[0]
                        self.remove(nextInd)
                        for i in temp:
                            self.insert(i)
                        nextInd = self.next(ind)
                        
                else:
                    temp1 = ind.complement(nextInd)[0]
                    temp2 = nextInd.complement(ind)[0]
                    
                    ind.common(nextInd)
                    if ind.hasAttribute():
                        ind.attribute.sum(nextInd.attribute)
                    self.remove(nextInd)
                    ind = temp1
                    self.insert(temp1)
                    self.insert(temp2)
                    nextInd = self.next(temp1)

            elif ind.compareValue(nextInd) == 1:
                toBeRemoved = nextInd                
                nextInd = self.next(nextInd)
                disorganized.append(toBeRemoved)
                self.remove(toBeRemoved)
            else:
                ind = nextInd
                nextInd = self.next(nextInd)  
        for ins in disorganized:
            self.insert(ins)
            
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()    
                
        ind = self.first()
        while ind is not None:
            ind = self.oneStepForward(ind)
        return
    
    
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        
        for i in other.elements:
            self.insert(i)
                
        self.maximalize()
        other.maximalize()
        return

    """
    difference of two forms
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.difference(formB)
      -->formA: a
    """
    def difference(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        
        self.maximalize()
        other.maximalize()
            
            
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:                            
                        otherInd = other.next(otherInd)
                    else:                             
                        thisInd =self.next(thisInd)       
                elif thisInd.compareInterval(otherInd) == 0:
                    if thisInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        toBeRemovedInd = thisInd
                        thisInd = self.next(thisInd)
                        self.remove(toBeRemovedInd)
                    otherInd = other.next(otherInd)
                elif thisInd.contains(otherInd):
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                        if thisInd.hasAttribute():
                            thisInd.common(otherInd)
                            for i in temp:
                                self.insert(i)
                            thisInd = self.next(thisInd)
                        else:
                            self.remove(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisInd = temp[0]
                    elif not thisInd.hasAttribute():
                        tobeRemoved = thisInd
                        self.remove(tobeRemoved)
                        for i in temp:
                            self.insert(i)
                        thisInd = temp[0]
   
                elif otherInd.contains(thisInd):
                    if thisInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        tobeRemoved = thisInd
                        thisInd = self.next(thisInd)
                        self.remove(tobeRemoved)
                    else:
                        thisInd = self.next(thisInd)
                elif thisInd.compareInterval(otherInd) == -1:   
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute():
                        temp[0].addAttribute(thisInd.attribute)
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        thisInd.common(otherInd)
                        self.insert(temp[0])
                        thisInd = self.next(thisInd)
                    else:
                        self.remove(thisInd)
                        self.insert(temp[0])
                        thisInd = temp[0]
                else:
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute():
                        temp[0].addAttribute(thisInd.attribute)
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        thisInd.common(otherInd)
                    else:
                        self.remove(thisInd)
                    self.insert(temp[0])
                    thisInd = temp[0]           
            elif thisInd.compareValue(otherInd) == 1: 
                otherInd = other.next(otherInd)
            else:                      
                thisInd =self.next(thisInd)    
        ind = self.first()
        while ind is not None:
            ind = self.oneStepForward(ind)
        return
        
    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    def product(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisInd = self.first()
        nextInd = self.next(thisInd)
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:                
                        otherInd = other.next(otherInd)
                    else:
                        self.remove(thisInd)
                        thisInd = nextInd
                        nextInd = self.next(thisInd)
                else:
                    temp = thisInd.duplicateInd()
                    temp.common(otherInd)
                    if thisInd.hasAttribute() and otherInd.hasAttribute():
                        temp.attribute.product(otherInd.attribute)
                        if temp.hasAttribute():
                            self.insert(temp)
                    else:
                        self.insert(temp)
                        
                    if thisInd.compareEnd(otherInd) < 1:
                        self.remove(thisInd)
                        thisInd = nextInd
                        nextInd = self.next(thisInd)
                    else:
                        otherInd = other.next(otherInd)
  
            elif thisInd.compareValue(otherInd) == 1: 
                otherInd = other.next(otherInd)
            else:         
                toBeRemovedInd = thisInd
                thisInd = self.next(thisInd)
                self.remove(toBeRemovedInd)
        while thisInd is not None:   
            self.remove(thisInd)
            thisInd = nextInd
            nextInd = self.next(thisInd)
        return

    """
    symdifference of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.symdifference(formB)
      -->formA: a,       d
    """
    def symdifference(self,other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherInd = other.next(otherInd)
                    else:         
                        thisInd = self.next(thisInd)    
                elif thisInd.compareInterval(otherInd) == 0:
                    thisInd = self.next(thisInd)    
                    otherInd = other.next(otherInd)
                else:         
                    temp1 = thisInd.complement(otherInd)
                    temp2 = otherInd.complement(thisInd)
                    temp3 = otherInd
                    temp4 = thisInd
                    if thisInd.compareEnd(otherInd) >= 0:
                        otherInd = other.next(otherInd)
                    elif thisInd.compareEnd(otherInd) <= 0:
                        thisInd = self.next(thisInd)    
                    temp3.common(temp4)
                    temp4.common(temp3)
                    for i in temp1:
                        self.insert(i)
                    for i in temp2:
                        other.insert(i)
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)    
                        
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        self.insert(otherInd)                            
                        otherInd = other.next(otherInd)
                    else:                             
                        thisInd = self.next(thisInd)    
                elif thisInd.compareInterval(otherInd) == 0:
                    #if has attribute, calculate difference
                    if thisInd.hasAttribute() or otherInd.hasAttribute() :
                        thisInd.attribute.symdifference(otherInd.attribute)
                        if not thisInd.hasAttribute():
                            toBeRemoved = thisInd
                            thisInd = self.next(thisInd)    
                            self.remove(toBeRemoved)
                        else:
                            thisInd = self.next(thisInd)    
                        otherInd = other.next(otherInd)
                    else:
                        toBeRemoved = thisInd
                        thisInd = self.next(thisInd)    
                        self.remove(toBeRemoved)
                        otherInd = other.next(otherInd)
                
            elif thisInd.compareValue(otherInd) == 1:
                self.insert(otherInd)                            
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)    
        while otherInd is not None:
            self.insert(otherInd)                            
            otherInd = other.next(otherInd)
            
        ind = self.first()
        while ind is not None:
            ind = self.oneStepForward(ind)
        ind = other.first()
        while ind is not None:
            ind = other.oneStepForward(ind)

        return 

    """
    partition of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
         
        _common = formA.partition(formB)
      -->formA: a
         formB:          d
         common:   b, c
    """
    def partition(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        

        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherInd = other.next(otherInd)
                    else:         
                        thisInd = self.next(thisInd)    
                elif thisInd.compareInterval(otherInd) == 0:
                    thisInd = self.next(thisInd)    
                    otherInd = other.next(otherInd)
                else:         
                    temp1 = thisInd.complement(otherInd)
                    temp2 = otherInd.complement(thisInd)
                    temp3 = otherInd
                    temp4 = thisInd
                    if thisInd.compareEnd(otherInd) >= 0:
                        otherInd = other.next(otherInd)
                    elif thisInd.compareEnd(otherInd) <= 0:
                        thisInd = self.next(thisInd)    
                    temp3.common(temp4)
                    temp4.common(temp3)
                    for i in temp1:
                        self.insert(i)
                    for i in temp2:
                        other.insert(i)
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)    
                
        
        common = interval(self.sort)
        
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:      
                        otherInd = other.next(otherInd)
                    else:                             
                        thisInd =self.next(thisInd)        
                elif thisInd.compareInterval(otherInd) == 0:
                    
                    attForm = otherInd.duplicateInd()
                    
                    if attForm.hasAttribute():
                        attForm.attribute.product(thisInd.attribute)
                        
                    if not (not attForm.hasAttribute() and (thisInd.hasAttribute() or otherInd.hasAttribute())):
                        common.insert(attForm)    
                    
                        if thisInd.hasAttribute():
                            thisInd.attribute.difference(attForm.attribute)
                        if otherInd.hasAttribute():
                            otherInd.attribute.difference(attForm.attribute)
                        
                        if not thisInd.hasAttribute():
                            toBeRemovedInd = thisInd
                            thisInd = self.next(thisInd)
                            self.remove(toBeRemovedInd)
                        else:
                            thisInd = self.next(thisInd)
                            
                        if not otherInd.hasAttribute():
                            toBeRemovedInd = otherInd
                            otherInd = other.next(otherInd)
                            other.remove(toBeRemovedInd)
                        else:
                            otherInd = other.next(otherInd)
                    else:
                        thisInd = self.next(thisInd)
                        otherInd = other.next(otherInd)
                        
            elif thisInd.compareValue(otherInd) == 1:      
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)       
        
        ind = self.first()
        while ind is not None:
            ind = self.oneStepForward(ind)
        ind = other.first()
        while ind is not None:
            ind = other.oneStepForward(ind)
        return common
        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        thisInd = self.first()
        otherInd = other.first()
        for thisInd in self.elements:
            check = False
            for otherInd in other.elements:
                if thisInd == otherInd:
                    check = True
                    if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                        return False
                elif otherInd.contains(thisInd):
                    check = True
                    if otherInd.hasAttribute() and not otherInd.attribute.partOf(thisInd.attribute):
                        return False
            if check == False:
                return False
        return True
