# -*- coding: utf-8 -*-
"""
Created on Thu Dec  1 10:36:10 2016

@author: akibdpt
"""
from ..planeSegment import planeSegment2D
from ..lineSegment import lineSegment2D
from ..line import line2D
from ...struct.bSegment import bSegment
from ...form.coForm import coForm

class crossAlgebraPlane2D():
    
    @staticmethod
    def freeSelection(vectorList, form):
        plane = planeSegment2D(form.sort, vectorList)
        return crossAlgebraPlane2D.crossAlgebraForm(plane, form)
    
    @staticmethod
    def crossAlgebraForm(pl, form):
        newForm = form.sort.newForm()
        if form.type == 'form-metaForm':
            for i in form.elements:
                temp = crossAlgebraPlane2D.crossAlgebraForm(pl,i)
                if not temp.isEmpty():
                    newForm.insert(temp)
        else:
            temp = crossAlgebraPlane2D.assessForm(pl,form)
            if temp is not None:
                return temp
        return newForm
                
    
    @staticmethod
    def assessForm(pl, form):
        if len(form.elements) == 0:
            return form
        if form.elements[0].type == 'ind-planeSegment2D':
            return crossAlgebraPlane2D.plane_plane(pl,form)
        elif form.elements[0].type == 'ind-line2D':
            return crossAlgebraPlane2D.plane_line(pl,form)
        elif form.elements[0].type == 'ind-lineSegment2D':
            return crossAlgebraPlane2D.plane_lineSeg(pl,form)
        elif form.elements[0].type == 'ind-point2D':
            return crossAlgebraPlane2D.plane_point(pl,form)
        return form
    
    @staticmethod
    def plane_plane(pl, planeForm):
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = coForm(planeForm.sort)
        for i in planeForm.elements:
            two = coForm(planeForm.sort)
            two.insert(i)
            
            classes = planeSegment2D.classification(one, two)
            if classes.disjoint() or classes.touches():
                return None
            else:
                temp = coForm(planeForm.sort)
                i.construct(classes.productSet(), temp)
                if i.hasAttribute():
                    for j in temp.elements:
                        j.addAttribute(i.attribute)
                temp.move(result)
        return result
    
    @staticmethod
    def plane_line(pl, lineForm):
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = lineForm.sort.newForm()
        for line in lineForm.elements:
            intersectionList = []
            for i in pl.boundaries:
                for j in i.canonical:
                    pLine = line2D(line.sort, j.start, j.end)
                    if pLine.intersect(line):
                        intersection = line.intersection(pLine)
                    else:
                        continue
                    if intersection >= j.start and intersection <= j.end:
                        intersectionList.append(intersection)
                        
            if len(intersectionList) == 0:
                continue
            else:
                result.insert(line)
        result.maximalize()
        return result
    
    @staticmethod
    def plane_lineSeg(pl, lineSegForm):
        one = coForm(pl.sort)
        one.insert(pl)
        bSeg = []
        
        for lineSeg in lineSegForm.elements:
            bSeg.append(bSegment(lineSeg.start,lineSeg.end,lineSeg))
        classes = planeSegment2D.classification_line(one, bSeg, False)
        if classes.disjoint():
            return None
        else:
            result = lineSeg.sort.newForm()
            for i in classes.containedSet():
                temp = lineSegment2D(lineSeg.sort,i.start, i.end)
                if lineSegForm.sort.type == 'sort-attribute':
                    temp.addAttribute(i.assoc.attribute)

                result.insert(temp)
            result.maximalize()
            return result
    
    @staticmethod
    def plane_point(pl, pointForm):
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = pointForm.sort.newForm()
        
        bSeg = []
        for point in pointForm.elements:
            vec = None
            lenV = 0
            for i in pl.boundaries:
                stoploop = False
                for j in i.canonical:
                    pointVect = point.value
                    u = j.direction()
                    v = pointVect.subtract(j.root())
                    if u.parallel(v):
                        if pointVect >= j.start and pointVect <= j.end: 
                            result.insert(point)
                            stoploop = True
                            break 
                        else:
                            continue
                    newVec = j.root().add(u.scale(v))
                    newLenV = newVec.subtract(pointVect)
                    if vec is None or newLenV.squareLength() < lenV.squareLength():
                        vec = newVec
                        lenV = newLenV
                if stoploop:
                    break
            if not stoploop:
                bSeg.append(bSegment(vec,pointVect, point))
            
        classes = planeSegment2D.classification_line(one, bSeg, False)
        for i in classes.containedSet():
            result.insert(i.assoc)
        result.maximalize()
        return result
        
    
"""
class crossAlgebra():
    
    @staticmethod
    def crossAlgebraForm(pl, form):
        newForm = form.sort.newForm()
        if form.type == 'form-metaForm':
            for i in form.elements:
                temp = crossAlgebra.crossAlgebraForm(pl,i)
                if not temp.isEmpty():
                    newForm.insert(temp)
        else:
            for i in form.elements:
                temp = crossAlgebra.assessInd(pl,i)
                if temp is not None:
                    for j in temp.elements:
                        if j is not None and not j.isNil():
                            newForm.insert(j)
        return newForm
                
    
    @staticmethod
    def assessInd(pl, ind):
        if ind.type == 'ind-planeSegment2D':
            return crossAlgebra.plane_plane(pl,ind)
        elif ind.type == 'ind-line2D':
            return crossAlgebra.plane_line(pl,ind)
        elif ind.type == 'ind-lineSegment2D':
            return crossAlgebra.plane_lineSeg(pl,ind)
        elif ind.type == 'ind-point2D':
            return crossAlgebra.plane_point(pl,ind)
        return ind
    
    @staticmethod
    def plane_plane(pl, plane):
        one = coForm(pl.sort)
        one.insert(pl)
        
        two = coForm(plane.sort)
        two.insert(plane)
        
        classes = planeSegment2D.classification(one, two)
        if classes.disjoint() or classes.touches():
            return None
        else:
            result = coForm(plane.sort)
            plane.construct(classes.productSet(), result)
            if plane.hasAttribute():
                for i in result.elements:
                    i.addAttribute(plane.attribute)
            return result
    
    @staticmethod
    def plane_line(pl, line):
        one = coForm(pl.sort)
        one.insert(pl)
        
        intersectionList = []
        for i in pl.boundaries:
            for j in i.canonical:
                pLine = line2D(line.sort, j.start, j.end)
                if pLine.intersect(line):
                    intersection = line.intersection(pLine)
                else:
                    continue
                if intersection >= j.start and intersection <= j.end:
                    intersectionList.append(intersection)
                    
        if len(intersectionList) == 0:
            return None
        if len(intersectionList) == 2:
            bSeg = bSegment(intersectionList[0],intersectionList[1])
        else:
            bSeg = None
            for i in range(len(intersectionList)):
                for j in range(i+1, len(intersectionList)):
                    temp = bSegment(intersectionList[i],intersectionList[j])
                    if bSeg is None or bSeg.squareLength() < temp.squareLength():
                        bSeg = temp
                
        classes = planeSegment2D.classification_line(one, bSeg, False)
        if classes.disjoint():
            return None
        else:
            result = line.sort.newForm()
            result.insert(line)
            return result
    
    @staticmethod
    def plane_lineSeg(pl, lineSeg):
        one = coForm(pl.sort)
        one.insert(pl)
        
        bSeg = bSegment(lineSeg.start,lineSeg.end)
        
        classes = planeSegment2D.classification_line(one, bSeg, False)
        if classes.disjoint():
            return None
        else:
            result = lineSeg.sort.newForm()
            a = classes.containedSet()
            for i in a:
                temp = lineSegment2D(lineSeg.sort,i.start, i.end)
                if lineSeg.hasAttribute():
                    temp.addAttribute(lineSeg.attribute)
                result.insert(temp)
            return result
    
    @staticmethod
    def plane_point(pl, point):
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = point.sort.newForm()
        vec = None
        lenV = 0
        for i in pl.boundaries:
            for j in i.canonical:
                pointVect = point.value
                u = j.direction()
                v = pointVect.subtract(j.root())
                if u.parallel(v):
                    if pointVect >= j.start and pointVect <= j.end: 
                        result.insert(point)
                        return result 
                    else:
                        continue
                newVec = j.root().add(u.scale(v))
                newLenV = newVec.subtract(pointVect)
                if vec is None or newLenV.squareLength() < lenV.squareLength():
                    vec = newVec
                    lenV = newLenV
        bSeg = bSegment(vec,pointVect)
            
        classes = planeSegment2D.classification_line(one, bSeg, False)
        if classes.disjoint():
            return None
        else:
            result.insert(point)
            return result
        
    
        
"""