# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 10:26:35 2016

@author: akibdpt
"""
from copy import deepcopy, copy

from ..element import element
from ..sort.attributeSort import attributeSort

"""
 * An individual is the representational unit of a {Sort}.
 * An individual specifies a data element with a single value;
 * individuals may be collected into a {Form}.
 * The value an individual may hold is dependent on the sort it belongs to.
 * Every {PrimitiveSort} is defined by its characteristic individual,
 * i.e., a representational abstraction of all individuals of the sort,
 * and by a description of the algebraic behavior of its forms. The algebraic
 * behavior of forms of individuals is derived from a partial order relationship
 * defined on the individuals and forms of a sort. This partial order
 * relationship defines the following operations and relations on individuals:
 * an individual may contain another individual, this is always the case
 * if they are equal; otherwise, two individuals may be disjoint; if two
 * individuals are not disjoint, they combine into another individual
 * that contains both individuals; two individuals that are not disjoint may
 * have a common part that is contained by both individuals; and an
 * individual may have a complement part wrt another individual, this is
 * always the case of the other individual does not contain this individual.
 * An individual may have an associate individual (under the object-attribute
 * relationship). This is the associate of the form this individual belongs to.
 * 
 * The Individual class extends on the Element class. It is an
 * abstract class specifying the common framework for all characteristic
 * individual classes. An instance of any characteristic individual class
 * defines an individual. Every individual belongs to a {PrimitiveSort};
 * the individual is necessarily an instance of the sort's characteristic
 * individual class. An individual is an Element with an optional
 * reference {Key} and an optional attribute Form.
 * The reference key serves to relate individuals. The optional attribute form
 * must be consistent with the sort this individual belongs to.
"""
class individual(element):
    """
    initialize individual
    create an attribute form if the individual's sort is an attribute sort
    """
    def __init__(self, sort = None):
        super().__init__(sort)
        if sort.type == 'sort-attribute':
            self.attribute = self.sort.weight.newForm()
            self.attribute.assoc = self
        else: self.attribute = None
        self.key = 0
        self.referenced = False
        return
        
        #attributes methods

    def __str__(self):
        return self.type+'_'+str(self.value)
    """
    check if an individual has attribute
    """
    def hasAttribute(self):
        return self.attribute is not None and not self.attribute.isEmpty()
        
    """
    get the attribute of the individual
    """
    def getAttribute(self):
        if (self.attribute is None) and isinstance(self.sort, attributeSort):
            print('ok')
            self.attribute = self.sort.weight.newForm()
            self.attribute.assoc = self
        else:
            print('not ok')
        return self.attribute
        
    """
    add an element as attribute to this individual
    """
    def addAttribute(self, element):
        if 'form' in element.type:
            for i in element:
                if i is not None and not i.isNil():
                    self.attribute.insert(i)
        elif 'ind' in element.type:
            self.attribute.insert(element)
        
    """
    not in use
    """
    def setAttribute(self, element):
        raise NotImplementedError
        if self.sort.type == 'sort-primitive':
            if self.assoc is not None:
                print("Attribute form cannot be set if individual belongs to a form")
                return
            #
        else: 
            #if isinstance(element.sort, recursiveSort):
            #   sort = sort.instance()
            if not element.sort == self.sort.weight:
                print("Attribute of erroneous sort")
                return
        if isinstance(element, individual):
            self.attribute = self.sort.newForm()
            self.attribute.add
        element.assoc = self
        self.attribute = element
        return 
        
    
    """
    check equality
    """
    def __eq__(self, other):
        if other is None or not isinstance(other, individual):
            return False
        if self.sort == other.sort:
            return (self.attribute == other.attribute) and (self.key == other.key) and (self.referenced == other.referenced)
        return False
    
    def compareValue(self, other):
        raise NotImplementedError 
        
    def duplicateInd(self):
        raise NotImplementedError 
        
    def setMin(self):
        raise NotImplementedError 
        
    """
    check the individual for rule analysis and rule detection
    """
    def accept(self, visitor, direction = -1):
        if self.isNil():
            return
        #if direction > 0:
        #    raise KeyError('is this even used?')
        #    if visitor.visitAssoc(self) and self.assoc is not None:
        #        self.assoc.accept(visitor, direction)
        #elif direction < 0:
        if visitor.visitEnter(self) and self.hasAttribute():
            self.attribute.accept(visitor, direction)
        #else:
        #    visitor.visitAssoc(self)
        #visitor.visitLeave(self)
    
    """
    return the codescriptor of the individual, usually the individual itself 
    except for line segment which has the line containing it as codescriptor
    """
    def codescriptor(self) :
        return self.duplicateInd()
    
    """
    editing the individual with a certain transformation (editor)
    """
    def edit(self, editor):
        if self.isNil():
            raise AttributeError('Individual may not be nil')
        if editor is not None:
            result = self.editInd(editor)
            if self.hasAttribute():
                result.sort = self.sort
                result.addAttribute(self.attribute.edit(editor))
    
        else:
            result = self.duplicateInd()
            if self.hasAttribute():
                attr = self.attribute.edit(None)
                attr.assoc = result
                result.addAttribute(attr)
        return result
        
    """
    editing the individual with a certain transformation (editor)
    """
    def editInd(self, editor):
        if editor.type == 'transformSet':
            return self.transformInd(editor)
        elif editor.type == 'paramSet':
            return self.evaluateInd(editor)
        return self.duplicateInd()
        
    def transformInd(self, transformSet):
        return copy(self)
        
    def evaluateInd(self, paramSet):
        raise NotImplementedError
        
    """
    converting the individuals that use coordinate vector into using float vectors  
    """
    def convertFloat(self):
        return copy(self)
            
    """
    check if the individual has the same value as element after converting the
    individual into float
    """
    def findFloatValue(self, element):
        if self.value == element.value:
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    """
    check if the individual belongs to a certain form after transformation with
    a transform set
    """
    def checkTransform(self, transform, compareForm):
        if compareForm.containsInd(self):
            return True
        return False
        
    """
    check if this individual contains another individual
    """
    def containsInd(self, ind):
        if self != ind:
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.partOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    def clear(self):
        if self.attribute is not None:
            self.attribute.clear()

    
    #------------------------------------------------------------
"""    
    def hashcode(self):
        return
    def compareTo(self):
        return
    def coDescriptor(self):
        return 
    def convertInd(self):
        return 
    def clear(self):
        return 
    def toString(self):
        return 
    def toStringInd(self):
        return 
    def parse(self):
        return 
    def accept(self):
        return 
    def parseIndividual(self):
        return 
    def transformInd(self):
        return 
    def editInd(self):
        return   
    def edit(self):
        return    
    def evaluateInd(self):
        return    
""" 


