# -*- coding: utf-8 -*-
"""
Created on Fri Nov  4 16:22:16 2016

@author: akibdpt
"""
from .individual import individual
from ..struct.vector import vector3D

"""
should only be used for 3D planes
"""
class planar(individual):
    def __init__(self, sort, boundary):
        super().__init__(sort)
        self.normal = boundary[0].product(boundary[1]).unit()
        
        self.scalar = boundary[2].scalar(boundary[2])
        self.root = self.normal.scale(self.scalar)
        
        if self.root.parallel(vector3D(0,0,1)):
            self.u = vector3D(1,0,0)
            self.v = vector3D(0,1,0)
        else:
            self.u = self.normal.product(vector3D(0,0,1)).unit()
            self.v = self.normal.product(self.u).unit()
        
    def __str__(self):
        return 'normal: '+str(self.normal)+'root: '+str(self.root)

    def isNil(self):
        return self.normal.isZero()
        
    def direction(self, coords):
        return self.u.scale(coords[0]).add(self.v.scale(coords[1]))
        
    def position(self, coords):
        return self.root.add(self.direction(coords))
        
    def localize(self, point):
        diff = point.subtract(self.root)
        return [self.u.scalar(diff), self.v.scalar(diff)]
        
    def scalar(self, direction):
        return [self.u.scalar(direction), self.v.scalar(direction)]
        
    def projection(self, point):
        return self.position(self.localize(point))
        
    @staticmethod
    def compareAngle(center, first, second):
        return first.product(second).add(center.product(first.subtract(second))).sign()
        
    def compareValue(self, other):
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        return self.scalar.compareValue(other.scalar)
        
    def parallel(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither planar may be nil')
        return self.normal == other.normal
    
    def perpendicular(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither planar may be nil')
        return self.normal.perpendicular(other.normal)

