# -*- coding: utf-8 -*-
"""
Created on Mon Nov  7 10:28:01 2016

@author: akibdpt
"""
from .individual import individual
from ..struct.vector import vector3D
from ..transform.affineTransform import affineTransform3D

class plane2D(individual):
    def __init__(self, sort):
        super().__init__(sort)
        self.type = 'ind-plane2D'
    
    def __str__(self):
        return 'plane'
        
    def __eq__(self, other):
        if other.type == 'ind-plane2D':
            return True
        return False
    
    def compareValue(self, other):
        if other.type == 'ind-plane2D':
            return 0
        return False
        
    def findFloatValue(self, element):
        if element.type == 'ind-plane2D':
            return True
        return False
        
    def parallel(self, other):
        if other.type == 'ind-plane2D':
            return True
        return False
        
class plane3D(individual):
    def __init__(self, sort, boundary):
        super().__init__(sort)
        self.type = 'ind-plane3D'
        self.normal = ((boundary[2].subtract(boundary[0])).product(
                        boundary[1].subtract(boundary[0]))).unit()
        
        self.scalar = boundary[2].scalar(boundary[2])
        self.root = self.normal.scale(self.scalar)
        
        a = self.root.__class__(0,0,1)
        if self.root.parallel(a):
            self.u = vector3D(1,0,0)
            self.v = vector3D(0,1,0)
        else:
            self.u = self.normal.product(self.normal.__class__(0,0,1)).unit()
            self.v = self.normal.product(self.u).unit()
        
    def __str__(self):
        return 'normal: '+str(self.normal)+'root: '+str(self.root)
        
    def __eq__(self, other):
        if other.type != 'ind-plane3D':
            return False
        return self.normal == other.normal and self.scale == other.scale
        
    def compareValue(self, other):
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        return self.scalar.compareValue(other.scalar)
                
    def isNil(self):
        return self.normal.isZero()
        
    def direction(self, coords):
        return self.u.scale(coords[0]).add(self.v.scale(coords[1]))
        
    def position(self, coords):
        return self.root.add(self.direction(coords))
        
    def localize(self, point):
        diff = point.subtract(self.root)
        return [self.u.scalar(diff), self.v.scalar(diff)]
        
    def scalar(self, direction):
        return [self.u.scalar(direction), self.v.scalar(direction)]
        
    def projection(self, point):
        return self.position(self.localize(point))
        
    @staticmethod
    def compareAngle(center, first, second):
        return first.product(second).add(center.product(first.subtract(second))).sign()
        
    def parallel(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither planar may be nil')
        return self.normal == other.normal
    
    def perpendicular(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither planar may be nil')
        return self.normal.perpendicular(other.normal)
            
    def findFloatValue(self, element):
        if element.type == 'ind-plane3D':
            return False
        if (self.normal.convertFloat() == element.normal and
            self.scale.convertFloat() == element.scale):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        position0 = self.root.transform(affine);
        position1 = self.root.add(self.u).transform(affine)      
        position2 = self.root.add(self.v).transform(affine)      
        return plane3D(self.sort, [position0,position1,position2])
        
    def intersection(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        if self.parallel(other):
            raise ValueError('The two planes are parallel to one another')
        prod = self.normal().product(other.normal())
        result = []
        result.append(prod.unit())
        result.append(self.root().add(prod.product(self.normal()).scale(self.scalar())))
        return result #return a set of two vectors, 1st vector: direction vector,
                      #                             2nd vector: root vector
