# -*- coding: utf-8 -*-
"""
Created on Mon Nov  7 10:30:24 2016

@author: akibdpt
"""
from copy import  copy as cp, deepcopy as copy
from .plane import plane2D
from ..ind.lineSegment import lineSegment2D
from ..ind.line import line2D
from ..struct.bSegment import bSegment
from ..struct.classes import classes as cls, lineClasses
from ..struct.splayTree import splayTree
from ..struct.splayable import sSeg, sRef, vertex
from ..form.coForm import coForm
from ..transform.affineTransform import affineTransform2D

class planeSegment2D(plane2D):
    
    """
    boundr: list of vectors or a boundary object (specified by boundaryList)
    boundaryList: False: boundary is list of vectors, 
                  True: boundary is a boundary object
    """
    def __init__(self, sort, boundr, boundaryList = False):
        super().__init__(sort)
        self.type = 'ind-planeSegment2D'
        if boundaryList: #just make self.boundaries = boundary if it is a boundary object
            self.boundaries = boundr
        else: # boundary is a list of vectors:
            #create a blank boundary object
            self.boundaries = [boundary()]
            #add line segments defining the boundary: 1st vector + 2nd vector = 1st line
            #                                         2nd vector + 3rd vector = 2nd line, etc
            for i in range(1,len(boundr)):
                self.boundaries[0].add(bSegment(boundr[i-1],boundr[i]))
            #add the last line segment defining the boundary: last vector + 1st vector = last line
            self.boundaries[0].add(bSegment(boundr[-1],boundr[0]))
            """???????????????????????"""
            classes = self.classify(None)
            result = coForm(self.sort)
            self.construct(classes.sumSet(), result)
            self.boundaries = result.first().boundaries
            """???????????????????????"""
            
            
    def __eq__(self, other):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other object must be a 2D plane segment')
        if len(self.boundaries) != len(other.boundaries):
            return False
        for i in self.boundaries:
            check = False
            for j in other.boundaries:
                if i == j:
                    check = True
            if check == False:
                return False
        return True

    
    def __str__(self):
        string = 'Plane Segment:\n'
        for i in self.boundaries:
            for j in i.canonical:
                string+= str(j)+'\n'
        return string
                        
    def isNil(self):
        if len(self.boundaries) == 0:
            return True
        for i in self.boundaries:
            if len(i.canonical) == 0 and len(i.polygonal) == 0:
                return True
        return False
        
    def printPlaneSegment(self, string = ''):
        print(string,'_ plane Segment:')
        for j in range(len(self.boundaries)):
            print(string,'__ boundary',j+1)
            """
            for i in self.boundaries[j]:
                print(string,'___ +segment:',i)
            """
            for i in self.boundaries[j].polygonal:
                print(string,'    _position:',i)
    
    def polygon(self):
        if len(self.boundaries) == 1:
            return self.boundaries[0].polygonal
        polygon = []
        polygon += self.boundaries[0].polygonal
        polygon.append(polygon[0])
        polygon += self.boundaries[1].polygonal
        polygon.append(self.boundaries[1].polygonal[0])
        return iter(polygon)
            
    def codescriptor(self):
        return plane2D(self.sort)
        
    def compareAreal(self, other):
        if other.type != 'ind-planeSegment2D':
            raise Exception('other is not a plane segment')
        bounds = other.boundaries
        c = len(self.boundaries) - len(bounds)
        if c != 0:
            return c/abs(c)
        for i in range(min(len(self.boundaries), len(bounds))):
            c = len(self.boundaries[i]) - len(bounds[i])
            if c != 0:
                return c/abs(c)
            for j in range(min(len(self.boundaries[i]), len(bounds[i]))):
                c = self.boundaries[i].canonical[j].compareValue(bounds[i].canonical[j])
                if (c != 0):
                    return c
        return 0
        
    def compareValue(self, other):
        if other.type == 'ind-planeSegment2D':
            return True
        else:
            return self.codescriptor().compareValue(other.codescriptor())
            
    def coEquals(self, other):
        if other.type == 'ind-planeSegment2D':
            return True
        else:
            return self.codescriptor().compareValue(other.codescriptor()) == 0
            
    """
     * Classifies the boundary segments of two coequal plane-segments,
     * each contained within a co-form, wrt each other into the categories of
     * inner and outer boundary segments for each plane-segment and the
     * categories of same-shared and oppositely-shared boundary segments.
    """        
    @staticmethod
    def classification(thisForm, otherForm, printing = False):
        """********"""
        if printing:
            print('\n\n_____________________________________________________________')
        """********"""
        classes = cls()
        # create the status of the sweepline as a splay tree
        # create a splayTree with two boundary segments: 1 from negative y infinity (0, -inf) to positive x infinity(inf, 0)
        #                                                1 from positive y infinity (0,  inf) to positive x infinity(inf, 0)
        """
                           \ |      /
                            \|     /
                      (0,inf)\    /
                             |\  /
                 ____________|_\/____________
                             | /\ (inf, 0)
                             |/  \
                     (0,-inf)/    \
                            /|     \
                           / |      \
                           
           tree:
                           |
                   (0,-inf),(inf,0)   <--- root
                    /            \
                  None     (0,inf),(inf,0)     
        """
        sweepline = splayTree()
        neginf = sSeg(bSegment(bSegment.NEG_Y_INF, bSegment.POS_X_INF))
        posinf = sSeg(bSegment(bSegment.POS_Y_INF, bSegment.POS_X_INF))
        neginf.alignedWithOne = False
        posinf.alignedWithOne = True
        sweepline.initSentinelTree(neginf, posinf)
        
        # create the sweep schedule of transition points as a splay tree
        transitions = splayTree()
        memberOfOther = False
        coform = thisForm
        while coform is not None:
            #for each of the bsegment in each of the boundary in the coform:
            for segment in coform.elements:
                for i in range(len(segment.boundaries)):
                    for val in segment.boundaries[i]:
                        #create a new splay segment of the bSegment with membership: False for thisForm, True for otherForm
                        interval = sSeg(val)
                        interval.membership = memberOfOther
                        #add start and end position of the splay segment into transition splay tree
                        transitions.insert(sRef.startPosition(interval))
                        transitions.insert(sRef.endPosition(interval))
            #switch to check otherform if currently checking thisForm, else break while loop
            if coform is thisForm:
                coform = otherForm
            else:
                coform = None
            memberOfOther = True
            
        # sweep across each of the transition points
        while not transitions.empty():
            #################################
            """********"""
            if printing:
                print('\n===transitions===')
                transitions.printPredSucc()
                print('---sweepline---')
                sweepline.printPredSucc()
                print('---------------')
            """********"""
            ##################################
            shared = False
            point = transitions.extractMin()
            # locate point within plane slice defined by sweepline status
            if printing:
                print('__ point:', point)
            node = sweepline.find(point.position)
            #splay compare the point with the bSegment of the tree node:
            #                       side == 0 --> point is colinear with the segment
            if printing:
                print('__ node:',node.content)
            side = node.content.splayCompare(point.position) * -1
            ##################################
            """********"""
            if printing:
                print('__ side:', side)
            """********"""
            ##################################
            if side > 0:
                low = node
                high = node.succ
            elif side < 0:
                low = node.pred
                high = node
            else:
                # determine set of edges containing current transition point
                # and delete edges ending at transition point
                low = node.pred
                prev = low
                node = low.succ
                """********"""
                if printing:
                    print('    __low:', low.content)
                    print('    __node:', node.content)
                """********"""
                # while point is still colinear with node bSegment (either being an endpoint of the line segment or on the segment):
                while node.content.splayCompare(point.position) == 0:
                    interval = node.content
                    #print('    __interval:', interval)
                    #extract (remove) the node from sweepline if point is the end position of segment
                    if interval.segment.end == point.position:
                        if printing:
                            print("    __point is interval's end -> extract node")
                        sweepline.extract(node)
                    else: # if the point is on the line segment but not an end position --> split the segment into two
                        node.content = sSeg(bSegment(point.position, interval.segment.end),interval)
                        interval = sSeg(bSegment(interval.segment.start, point.position),interval)
                        if printing:
                            print('    __point not end point --> split segment')                        
                            print('      __node:    ', node.content)
                            print('      __interval:', interval)
                        prev = node
                        
                    # classify edge into one of the categories:
                    # inner, outer, same-shared or opposite-shared
                    """********"""
                    if printing:
                        print('       +++', interval.segment, prev.succ.content.segment)
                        print('       +++',shared, interval.segment.start == prev.succ.content.segment.start, interval.segment.direction().parallel(prev.succ.content.segment.direction()), interval.membership)
                    """********"""
                    if shared: #both plane segments contain this bSegment
                        """********"""
                        if printing:
                            print('       +++interval shared')
                        """********"""
                        if interval.alignedWithOne == interval.alignedWithOther:
                            #segment is align with both plane segments --> same shared
                            classes.addSameShared(interval.segment)
                        else:
                            #segment have different alignment with the two plane segments --> opposite shared
                            classes.addOppShared(interval.segment)
                        shared = False #reset shared for next segment 
                    #
                    elif (interval.segment.start == prev.succ.content.segment.start 
                            and interval.segment.direction().parallel(prev.succ.content.segment.direction())):
                        """********"""
                        if printing:
                            print('       @@ next -> shared')
                        """********"""
                        shared = True
                    elif interval.membership: #segment is a member of other plane segment only
                        """********"""
                        if printing:
                            print('       +++interval member of other')
                        """********"""
                        if interval.alignedWithOne: # align with this plane segment
                            classes.addInnerTwo(interval.segment) # --> I2
                        else: # does not a lign with this plane segment
                            classes.addOuterTwo(interval.segment) # --> O2
                    else: # segment is a member of this plane segment only
                        """********"""
                        if printing:
                            print('       +++interval member of self')
                        """********"""
                        if interval.alignedWithOther: # align with other plane segment
                            classes.addInnerOne(interval.segment) # --> I1
                        else:  # does not align with other plane segment
                            classes.addOuterOne(interval.segment) # --> O1
                    node = prev.succ 
                    """********"""
                    if printing:
                        print('    __node:    ', node.content)
                    """********"""
                high = node
                # reverse the order of all other edges in the set
                if prev != low: 
                    sweepline.reverse(low.succ, prev)
                
            #insert all edges with transition point as start point
            ref = point.position
            if point.isStartPosition():
                sweepline.insert(point.segment)
            while (not transitions.empty()) and ref == transitions.findMin().content.position:
                
                point = transitions.extractMin()
                if point.isStartPosition():
                    #insert edge if transition point of type start
                    sweepline.insert(point.segment)

            # update alignment information
            alignedWithOne = low.content.alignedWithOne
            alignedWithOther = low.content.alignedWithOther  
            """********"""
            if printing:
                print('== low:',low.content)
                print('== high:', high.content)
                print('++ low  align:',alignedWithOne, alignedWithOther)
                print('++ high align:',alignedWithOne, alignedWithOther)
            """********"""
            node = low.succ
            while node != high:
                if node.content.membership:
                    """********"""
                    if printing:
                        print('    node align: other')
                    """********"""
                    alignedWithOther = not alignedWithOther
                else:
                    """********"""
                    if printing:
                        print('    node align: one')
                    """********"""
                    alignedWithOne = not alignedWithOne
                node.content.alignedWithOne = alignedWithOne
                node.content.alignedWithOther = alignedWithOther
                node = node.succ
            if high.content.membership:
                """********"""
                if printing:
                    print('    high align: other__high')
                    """********"""
                alignedWithOther = not alignedWithOther
            else:
                """********"""
                if printing:
                    print('    high align: one__high')
                """********"""
                alignedWithOne = not alignedWithOne
            """********"""
            if printing:
                print('^^ ',high)
                print('^^ ',high.content.alignedWithOne, high.content.alignedWithOther)
                print('^^ ',alignedWithOne, alignedWithOther)
            """********"""
            if (high.content.alignedWithOne != alignedWithOne or
                high.content.alignedWithOther != alignedWithOther):
                    raise Exception('alignment error')
                    
            # check for intersection point between neighboring segment 'low'
            # and immediate successor
            """********"""
            if printing:
                print('** find intersection low:', low.content, low.succ.content)
            """********"""
            intersection = low.content.segment.intersection(low.succ.content.segment, bSegment.OPEN);
            if intersection is not None:
                transitions.insert(sRef.intersection(intersection))
            # idem for neighboring segment 'high' and immediate predecessor
            if high.pred != low:
                """********"""
                if printing:
                    print('** find intersection high:', high.content, high.pred.content)
                """********"""
                intersection = high.content.segment.intersection(high.pred.content.segment, bSegment.OPEN);
                if intersection is not None:
                    transitions.insert(sRef.intersection(intersection))
        if len(sweepline) > 2:
            print("Sweep status not empty")
        sweepline.purge()
        transitions.purge()
    
        return classes
            
    def classify(self, other):
        thisForm = coForm(self.sort)
        thisForm.insert(self)
        otherForm = coForm(self.sort)
        if other is not None:
            otherForm.insert(other)
        result = self.classification(thisForm, otherForm)
        thisForm.remove(self)
        if other is not None:
            otherForm.remove(other)
        return result



    @staticmethod
    def classification_line(thisForm, bS, printing = False):
        """********"""
        if printing:
            print('\n\n_____________________________________________________________')
        """********"""
        classes = lineClasses()
                    
        # create the status of the sweepline as a splay tree
        # create a splayTree with two boundary segments: 1 from negative y infinity (0, -inf) to positive x infinity(inf, 0)
        #                                                1 from positive y infinity (0,  inf) to positive x infinity(inf, 0)
        """
                           \ |      /
                            \|     /
                      (0,inf)\    /
                             |\  /
                 ____________|_\/____________
                             | /\ (inf, 0)
                             |/  \
                     (0,-inf)/    \
                            /|     \
                           / |      \
                           
           tree:
                           |
                   (0,-inf),(inf,0)   <--- root
                    /            \
                  None     (0,inf),(inf,0)     
        """
        sweepline = splayTree()
        neginf = sSeg(bSegment(bSegment.NEG_Y_INF, bSegment.POS_X_INF))
        neginf.alignedWithOne = False
        posinf = sSeg(bSegment(bSegment.POS_Y_INF, bSegment.POS_X_INF))
        posinf.alignedWithOne = True
        sweepline.initSentinelTree(neginf, posinf)
        
        # create the sweep schedule of transition points as a splay tree
        transitions = splayTree()
        
        #for each of the bsegment in each of the boundary in the coform:
        for segment in thisForm.elements:
            for i in range(len(segment.boundaries)):
                for val in segment.boundaries[i]:
                    #create a new splay segment of the bSegment with membership: False for thisForm, True for otherForm
                    interval = sSeg(bSegment(val.start, val.end))
                    interval.membership = False
                    #add start and end position of the splay segment into transition splay tree
                    transitions.insert(sRef.startPosition(interval))
                    transitions.insert(sRef.endPosition(interval))
            #switch to check otherform if currently checking thisForm, else break while loop
        
        for s in bS:
            interval = sSeg(s)
            interval.membership = True
            transitions.insert(sRef.startPosition(interval))
            transitions.insert(sRef.endPosition(interval))

        # sweep across each of the transition points
        while not transitions.empty():
            #################################
            """********"""
            if printing:
                print('\n===transitions===')
                transitions.printPredSucc()
                print('---sweepline---')
                sweepline.printPredSucc()
                print('---------------')
            """********"""
            ##################################
            shared = False
            point = transitions.extractMin()
            # locate point within plane slice defined by sweepline status
            node = sweepline.find(point.position)
            #splay compare the point with the bSegment of the tree node:
            #                       side == 0 --> point is colinear with the segment
            side = node.content.splayCompare(point.position) * -1
            ##################################
            """********"""
            if printing:
                print('__ point:', point)
                print('__ node :', node.content)
                print('__ side :', side)
            """********"""            
            
            ##################################
            if side > 0:
                low = node
                high = node.succ
            elif side < 0:
                low = node.pred
                high = node
            else:
                # determine set of edges containing current transition point
                # and delete edges ending at transition point
                low = node.pred
                prev = low
                node = low.succ
                """********"""
                if printing:
                    print('    __low:', low.content)
                    print('    __node:', node.content)
                """********"""
                # while point is still colinear with node bSegment (either being an endpoint of the line segment or on the segment):
                while node.content.splayCompare(point.position) == 0:
                    interval = node.content
                    #print('    __interval:', interval)
                    #extract (remove) the node from sweepline if point is the end position of segment
                    if interval.segment.end == point.position:
                        if printing:
                            print("    __point is interval's end -> extract node")
                        sweepline.extract(node)
                    else: # if the point is on the line segment but not an end position --> split the segment into two
                        node.content = sSeg(bSegment(point.position, interval.segment.end, interval.segment.assoc),interval)
                        interval = sSeg(bSegment(interval.segment.start, point.position, interval.segment.assoc),interval)
                        if printing:
                            print('    __point not end point --> split segment')                        
                            print('      __node:    ', node.content)
                            print('      __interval:', interval)
                        prev = node
                        
                    # classify edge into one of the categories:
                    # inner, outer, same-shared or opposite-shared
                    """********"""
                    if printing:
                        print('       +++', interval.segment, prev.succ.content.segment)
                        print('       +++',shared, interval.segment.start == prev.succ.content.segment.start, interval.segment.direction().parallel(prev.succ.content.segment.direction()), interval.membership)
                    """********"""
                    if shared: #both plane segments contain this bSegment
                        """********"""
                        if printing:
                            print('       +++interval shared')
                        """********"""
                        if interval.segment.assoc is not None:
                            classes.addShared(interval.segment)
                        shared = False #reset shared for next segment 
                        
                    #
                    elif (interval.segment.start == prev.succ.content.segment.start 
                            and interval.segment.direction().parallel(prev.succ.content.segment.direction())):
                        """********"""
                        if printing:
                            print('       @@ next -> shared')
                        """********"""
                        shared = True
                        if interval.segment.assoc is not None:
                            classes.addShared(interval.segment)
                    elif interval.membership: #segment is a member of other plane segment only
                        """********"""
                        if printing:
                            print('       +++interval member of other')
                        """********"""
                        if interval.alignedWithOne: # align with this plane segment
                            classes.addInner(interval.segment) # --> I2
                        else: # does not a lign with this plane segment
                            classes.addOuter(interval.segment) # --> O2
                    node = prev.succ 
                    """********"""
                    if printing:
                        print('    __node:    ', node.content)
                    """********"""
                high = node
                # reverse the order of all other edges in the set
                if prev != low: 
                    sweepline.reverse(low.succ, prev)
                
            #insert all edges with transition point as start point
            ref = point.position
            if point.isStartPosition():
                sweepline.insert(point.segment)
            while (not transitions.empty()) and ref == transitions.findMin().content.position:
                
                point = transitions.extractMin()
                if point.isStartPosition():
                    #insert edge if transition point of type start
                    sweepline.insert(point.segment)

            # update alignment information
            alignedWithOne = low.content.alignedWithOne
            """********"""
            if printing:
                print('== low:',low.content)
                print('== high:', high.content)
                print('++ low  align:',alignedWithOne)
                print('++ high align:',alignedWithOne)
            """********"""
            node = low.succ
            while node != high:
                if not node.content.membership:
                    """********"""
                    if printing:
                        print('    node align: one')
                    """********"""
                    alignedWithOne = not alignedWithOne
                node.content.alignedWithOne = alignedWithOne
                node = node.succ
            if not high.content.membership:
                """********"""
                if printing:
                    print('    high align: one__high')
                """********"""
                alignedWithOne = not alignedWithOne
            """********"""
            if printing:
                print('^^ ',high)
                print('^^ ',high.content.alignedWithOne)
                print('^^ ',alignedWithOne)
            """********"""
            if (high.content.alignedWithOne != alignedWithOne):
                    raise Exception('alignment error')
                    
            # check for intersection point between neighboring segment 'low'
            # and immediate successor
            """********"""
            if printing:
                print('** find intersection low:', low.content, low.succ.content)
            """********"""
            intersection = low.content.segment.intersection(low.succ.content.segment, bSegment.OPEN);
            if intersection is not None:
                transitions.insert(sRef.intersection(intersection))
            # idem for neighboring segment 'high' and immediate predecessor
            if high.pred != low:
                """********"""
                if printing:
                    print('** find intersection high:', high.content, high.pred.content)
                """********"""
                intersection = high.content.segment.intersection(high.pred.content.segment, bSegment.OPEN);
                if intersection is not None:
                    transitions.insert(sRef.intersection(intersection))
        if len(sweepline) > 2:
            print("Sweep status not empty")
        sweepline.purge()
        transitions.purge()
    
        return classes
            

















    
    @staticmethod
    def clockwiseAngle(r, p, q):
        return p.subtract(r).product(q.subtract(r)) < 0

    
    @staticmethod
    def counterclockwiseAngle(r, p, q):
        return p.subtract(r).product(q.subtract(r)) > 0
    
    """/
     * Finds among all adjacency boundary segments of a vertex, the <b>starting
     * segment</b> with the most clockwise polar angle (is ccw from 'below'
     * reference line). This segment is removed from the vertex's adjacency
     * boundary segments list.
    """
    @staticmethod
    def startingSegment(vertex):
        # assume first edge is possible continuation edge
        interval = vertex.edges[0]

        # find edge with most clockwise polar angle
        for i in vertex.edges:
            # if new edge clockwise wrt to current, then update current
            if planeSegment2D.clockwiseAngle(vertex.position, interval.end, i.end):
                interval = i
        vertex.edges.remove(interval)
        return interval
    
    """
     * Finds among all adjacency boundary segments of a vertex, the
     * <b>continuation segment</b> with the smallest clockwise polar angle from
     * the last visited boundary segment. Both segments are removed from the
     * vertex's adjacency boundary segments list.
    """
    @staticmethod
    def continuationSegment(vertex):
        # if there are only two edges left then one has to be the
        # last visited edge and the other the continuation edge
        if len(vertex.edges) == 2:
            edge = vertex.edges[0]
            if edge == vertex.lastVisited():
                edge = vertex.edges[1]
            vertex.edges.clear()
            return edge
            
        # find and remove last visited edge from vertex
        i = 0
        for i in vertex.edges:
            if i == vertex.lastVisited():
                vertex.edges.remove(i)
                break
        # determine reference point for calculation of edge angle
        refpos = vertex.lastVisited().start
        if refpos == vertex.position:
            refpos = vertex.lastVisited().end;

        # assume first edge is possible result
        oldpos = vertex.edges[0].start
        if oldpos == vertex.position:
            oldpos = vertex.edges[0].end
        cw = not planeSegment2D.counterclockwiseAngle(vertex.position, oldpos, refpos)

        # find edge with the smallest clockwise polar angle from reference edge
        index = 0
        for i in range(1, len(vertex.edges)):
            newpos = vertex.edges[i].start
            if newpos == vertex.position:
                newpos = vertex.edges[i].end
            t = not planeSegment2D.counterclockwiseAngle(vertex.position, newpos, refpos)
            r = planeSegment2D.clockwiseAngle(vertex.position, newpos, oldpos)
            # if new edge counterclockwise wrt to current
            # and clockwise wrt reference, then update current
            if ((cw or r) and not t) or (cw and r):
                index = i
                oldpos = newpos
                cw = t
        result = vertex.edges[index]
        vertex.edges.remove(vertex.edges[index])
        return result
    
    
    """
     * Returns whether the start of a boundary segment is inside a
     * given boundary.
    """
    @staticmethod
    def insideplaneboundary(segment, boundary, normal):
        left = 0
        inside = True
        
        # take a scanline coincident with the segment
        # and check for intersectionpoints with the boundary
        for n in range(len(boundary)):
            if not inside:
                break
            bound = boundary[n]
            for current in bound:
                if (segment.aligns(current) and current.start.compareValue(segment.start) != 1
                    and segment.start.compareValue(current.end) == -1):
                        return segment.alignedPositive == current.alignedPositive
                ipt = segment.intersection(current, bSegment.INFINITE)

                if ipt is not None:
                    if ipt == current.start:
                        val = normal.dotproduct(current.end) #xvdot(normal, copoint(current->Head), val);
                        if val.sign() < 0:
                            continue
                    elif ipt == current.end:
                        val = normal.dotproduct(current.start) #xvdot(normal, copoint(current->Tail), val);
                        if val.sign() < 0:
                            continue
                    # keep track of the # of intersections left of the tail
                    if ipt.compareValue(segment.start) != 1:
                        left+=1
            #if ((ptr->prev == NULL) == !(left % 2)) inside = FALSE;
            if n == 0 and left % 2 == 0: 
                inside = False 
        return inside
    
    
    """
     * <b>Determines the boundary type</b> of a plane-segment boundary wrt a
     * list of plane-segment boundaries as inner or outer. If the plane-segment
     * boundary is determined as inner wrt one of the plane-segment boundaries
     * in the list, then its outer boundary will be added as an inner boundary
     * (hole) to this plane-segment boundary, while its inner boundaries (holes)
     * will be added as (outer) boundaries to the list of plane-segment
     * boundaries; otherwise, it will be added to the list of plane-segment
     * boundaries.
     * @param boundaries the boundary (list) of a single plane segment;
     * this list is emptied
     * @param segments the list of plane segment boundaries (boundary lists);
     * this list is updated
    """
    @staticmethod
    def determineboundarytype (boundaries, segments):
        if len(segments) == 0:
            segment = []
            segment += boundaries
            segments.append(segment)
            boundaries.clear()
            return
        
        boundary = boundaries[0]
                
        refval = boundary.canonical[0] #refl = bounds->FirstEl->FirstEl;
        
        # a non-colinear point is sufficient to define the normal vector
        point = boundary.canonical[-1].start #pt = bounds->FirstEl->LastEl->Tail;
        
        
        if point == refval.start or point == refval.end:
            point = boundary.canonical[-1].end #pt = bounds->FirstEl->LastEl->Head;
        
        # compute the normal vector to the plane
        normalv = point.subtract(refval.start); #xvsub(copoint(pt), copoint(refl->Tail), normv);
        planev = refval.direction().product(normalv) #vprod(coline(refl), normv, planev);
        
        # determine comparator plane equation (normal vector)
        normalv.x =  refval.direction().y*planev*(-1) #vprod(coline(refl), planev, normv);
        normalv.y = refval.direction().x*planev #vprod(coline(refl), planev, normv);
        #Rational factor = normalv.dotProduct(refval.tail()); #vdot(normv, copoint(refl->Tail), normv->h);
        #factor = factor.negate(); #cneg(normv->h, normv->h);
        #normalv = normalv.scale(refval.start.w) #vscal(normv, refl->Tail->Wcoord);
        
        n = 0
        while n < len(segments):
            if planeSegment2D.insideplaneboundary(refval, segments[n], normalv):
                break
            n+=1
        
        if n < len(segments):
            # new boundary is an inner boundary
            segments[n].append(boundaries.pop(0)) #shappend_and_extract(ptr->el, bounds);
            # inner boundaries become outer boundaries
            while len(boundaries) != 0:
                bounds = []
                bounds.append(boundaries.pop(0))
                segments.append(bounds)
        else:
            segment = []
            segment += boundaries
            segments.append(segment)
            boundaries.clear()
        
    """
     * Extracts the polygonal boundaries that define the plane
     * segments corresponding to a list of (non-intersecting) boundary
     * segments. The list of boundary segments is emptied.
    """
    @staticmethod
    def extractSegments (intervals, maximal):
        segments = []
        boundaries = []
        normal = abs(intervals[0].direction().product(intervals[-1].direction()))
        
        # create a sorted set of endposition vectors
        positions = []
        for interval in intervals:
            positions.append(interval.start)
            positions.append(interval.end)
        positions.sort()
        # create a graph of vertices from the set of endposition vectors
        graph = []
        vertex.vertices.clear()
        ref = vertex(positions.pop(0))
        graph.append(ref)
        
        while len(positions)!=0: 
            vector = positions.pop(0)
            if ref.position!=vector:
                ref = vertex(vector)
                graph.append(ref)
            
            string = str(vector.x)+'|'+str(vector.y)
            vertex.vertices[string] = ref
        
        # creates two edge-halves for each line segment,
        # with an entry into the adjacency list of each boundary segment
        while len(intervals)!=0:
            interval = intervals.pop(0)
            
            stringStart = str(interval.start.x)+'|'+str(interval.start.y)
            stringEnd = str(interval.end.x)+'|'+str(interval.end.y)
            
            vertex.vertices[stringStart].edges.append(interval)
            vertex.vertices[stringEnd].edges.append(interval)
                
        cycle = []
        #ArrayList<BSegment> boundary = new ArrayList<BSegment>();
        bdr = boundary()
        while len(graph)!=0: 
            
            # first vertex is leftmost bottom vertex (always start position)
            vert = graph[0]
            vert.setPartOfCycle(True)
            bdr.add(vert.position)
                        
            # determine starting edge ccw from vertical down direction
            interval = planeSegment2D.startingSegment(vert)

            cycle.append(interval) # add interval into cycle
            
            refv = copy(interval.direction())
            refv.x = interval.direction().y*normal*(-1)
            refv.y = interval.direction().x*normal
            refv = refv.add(vert.position)
            alignedPositive = True
            
            if planeSegment2D.clockwiseAngle(vert.position, interval.end, refv):
                alignedPositive = False
                
            
            interval.alignedPositive = alignedPositive
            stringEnd = str(interval.end.x)+'|'+str(interval.end.y)
            vert = vertex.vertices[stringEnd]	# next vertex is always end position (since first vertex is always start position)

            vert.setLastVisited(interval)
            bdr.add(vert.position)
            while len(cycle) != 0 :
                                
                vert.setPartOfCycle(True)
                # determine next edge cw from current edge
                interval = planeSegment2D.continuationSegment(vert)
                cycle.append(interval)
                # next vert
                if interval.start == vert.position:
                    stringEnd = str(interval.end.x)+'|'+str(interval.end.y)
                    vert = vertex.vertices[stringEnd]
                    interval.alignedPositive = alignedPositive
                else:
                    stringStart = str(interval.start.x)+'|'+str(interval.start.y)
                    vert = vertex.vertices[stringStart]
                    interval.alignedPositive = not alignedPositive
                vert.setLastVisited(interval)
                if vert.isPartOfCycle():
                    # if vert is already in cycle then a polygon is found
                    # create polygon (current) as a set of lines
                    segment = cycle.pop(-1)
                    stringStart = str(segment.start.x)+'|'+str(segment.start.y)
                    stringEnd = str(segment.end.x)+'|'+str(segment.end.y)
            
                    vertex.vertices[stringStart].setPartOfCycle(False)
                    vertex.vertices[stringEnd].setPartOfCycle(False)
                    bdr.add(segment)
                    while True:
                        segment = cycle.pop(-1)
                        stringStart = str(segment.start.x)+'|'+str(segment.start.y)
                        stringEnd = str(segment.end.x)+'|'+str(segment.end.y)
                        tail = vertex.vertices[stringStart]
                        tail.setPartOfCycle(False)
                        head = vertex.vertices[stringEnd]
                        head.setPartOfCycle(False)
                        bdr.add(segment)
                        if len(cycle) == 0 or tail is vert or head is vert:
                            break
                    bdr.sort()
                    #if (!maximal) boundary.maximalize();
                    
                    if len(cycle) == 0:
                        boundaries.insert(0, bdr)
                        bdr = boundary()
                        
                        # add boundaries to list of polygons
                        planeSegment2D.determineboundarytype(boundaries, segments)
                        # remove edge from new vert
                        vert.edges.remove(interval)
                    else:
                        boundaries.append(bdr)
                        bdr = boundary()
                    
                else: 
                    bdr.add(vert.position)
            # find first vertex with non-empty adjacency list
            while len(graph) != 0 and len(graph[0].edges) == 0:
                graph.pop(0)
        
        if len(cycle) != 0:
            raise KeyError("Some vertices are left on the cycle\n")
        
        cycle.clear()
        vertex.vertices.clear()
        return segments


    """
     * <b>Constructs</b> a plane-segment from a collection of boundary segments.
     * This collection of boundary segments must correspond to a proper
     * plane-segment boundary (list).
     * @param intervals a <tt>TreeSet</tt> collection of boundary segments
     * defining a plane-segment
     * @param result a {@link CoForm} object to contain the resulting
     * plane-segment
    """
    def construct(self, intervals, result):
        if len(intervals) == 0:
            return
        segments = planeSegment2D.extractSegments(intervals, True)
        while len(segments) != 0:
            result.insert(planeSegment2D(self.sort, segments.pop(0), True))
        result.maximalize()

    
    def contains(self, other):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).contains()
    
    def touches(self, other):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).touches()

    def disjoint(self, other):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).disjoint()

    def aligns(self, other):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).aligns()

    def combine(self, other, result):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        classes = self.classify(other)
        if classes.disjoint():
            return False
        self.construct(classes.sumSet(), result)
        
    def common(self, other, result):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        classes = self.classify(other);
        if classes.disjoint() or classes.touches():
            return False
        self.construct(classes.productSet(), result)
        return True
    
    def complement(self, other, result):
        if other.type != 'ind-planeSegment2D':
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        classes = self.classify(other)
        if classes.disjoint() or classes.touches():
            result.insert(self)
            return True
        if classes.isContained():
            return False
        self.construct(classes.diffSet(), result)
        return True
    
    def transformInd(self, transforms):
        affine = transforms.retrieveTransform(affineTransform2D)
        if affine is None:
            return copy(self)
        bounds = []
        for i in self.boundaries:
            bounds.append(i.transform(affine))
        return planeSegment2D(self.sort, bounds, True)
        
    def duplicateInd(self):
        boundary = []
        for i in self.boundaries:
            bound = boundary()
            for j in i.canonical:
                bound.canonical.append(j)
            for j in i.polygonal:
                bound.polygonal.append(j)
            boundary.append(bound)
        return planeSegment2D(self.sort, boundary, True)
        
    """
    def accept(self, visitor, direction = -1):
        if self.isNil():
            return
        for i in self.boundaries:
            for j in i:
                lineSeg = lineSegment2D(self.sort, j.start, j.end)
                if self.hasAttribute():
                    lineSeg.addAttribute(self.attribute)
                lineSeg.accept(visitor, direction)
    """
                
    def convertFloat(self):
        bdrs = []
        for i in self.boundaries:
            bdr =  boundary()
            for j in i.canonical:
                bSeg = bSegment(j.start.convertFloat(),j.end.convertFloat())
                bdr.canonical.append(bSeg)
            for j in i.polygonal:
                bdr.polygonal.append(j.convertFloat())
            bdrs.append(bdr)
        result = planeSegment2D(self.sort, bdrs, True)
        if self.hasAttribute():
            result.addAttribute(self.attribute)
        return result
        
    def findFloatValue(self, element):
        if element.type == "ind-lineSegment2D":
            for i in self.boundaries:
                for j in i.canonical:
                    if j.start.floatValue == element.start and j.end.floatValue == element.end:
                        return lineSegment2D(self.sort, j.start, j.end)
        elif element.type == "ind-line2D":
            for i in self.boundaries:
                for j in i.canonical:
                    line = line2D(self.sort, j.start, j.end)
                    if line.findFloatValue(element) is not False:
                        return line
        elif element.type == "ind-planeSegment2D":
            for i in self.boundaries:
                check = False
                for j in element.boundaries:
                    if i.findFloatValue(j):
                        check = True
                if check == False:
                    return False
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False

    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform2D)
        if affine is None:
            tfPlaneSeg = self
        else:
            bdrs = []
            for i in self.boundaries:
                bdr =  boundary()
                for j in i.canonical:
                    bSeg = bSegment(j.start.transform(affine),j.end.transform(affine))
                    bdr.canonical.append(bSeg)
                for j in i.polygonal:
                    bdr.polygonal.append(j.transform(affine))
                bdrs.append(bdr)
            tfPlaneSeg = planeSegment2D(self.sort, bdrs, True)
            if self.hasAttribute():
                tfPlaneSeg.addAttribute(self.attribute)
        if compareForm.containsInd(tfPlaneSeg):
            return True
        return False
        
    def containsInd(self, ind):
        f1 = coForm(self.sort)
        f1.insert(self)
        f2 = coForm(ind.sort)
        f2.insert(ind)
        c = planeSegment2D.classification(f1,f2)
        if not c.contains():
            return False

        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.partOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
        
        
        
"""
convertInd()
toStringInd()
parse()

"""
class boundary():
    
    def __init__(self):
        self.canonical = []
        self.polygonal = []
    
    def __len__(self):
        return len(self.canonical)
    
    def __eq__(self, other):
        if len(self.canonical) != len(other.canonical):
            return False
        for i in self.canonical:
            check = False
            for j in other.canonical:
                if i == j:
                    check =True
            if check == False:
                return False
        return True
    
    def add(self, obj):
        if (obj.type == 'struct-vector2D' or  obj.type == 'struct-vector2Dfloat' or
            obj.type == 'struct-vector3D' or  obj.type == 'struct-vector3Dfloat'):
            self.polygonal.append(obj)
        else:
            self.canonical.append(obj)
    
    def merge(self, other):
        self.canonical += other.canonical
        self.polygonal += other.polygonal
        
    def sort(self):
        self.canonical.sort()
    
    def partOf(self, other):
        for i in self.canonical:
            check = False
            for j in other.canonical:
                if i == j or i.partOf(j):
                    check = True
            if check == False:
                return False
        return True
        
    def maximalize(self):
        polygonCount = {}
        for i in self.polygonal:
            polygonCount[str(i)] = [0,None]
        for i in self.canonical:
            if polygonCount[str(i.start)][0] == -1:
                pass
            elif polygonCount[str(i.start)][0] == 0:
                polygonCount[str(i.start)] = [1,i.end]
            elif polygonCount[str(i.start)][0] == 2:
                if i.start.colinear(i.end, polygonCount[str(i.start)][1]):
                    polygonCount[str(i.start)] = [3,None]
                else:
                    polygonCount[str(i.start)] = [-1,None]
            else:
                polygonCount[str(i.start)] = [-1,None]
            if polygonCount[str(i.end)][0] == -1:
                pass
            elif polygonCount[str(i.end)][0] == 0:
                polygonCount[str(i.end)] = [2,i.start]
            elif polygonCount[str(i.end)][0] == 1:
                if i.start.colinear(i.end, polygonCount[str(i.end)][1]):
                    polygonCount[str(i.end)] = [3,None]
                else:
                    polygonCount[str(i.end)] = [-1,None]
            else:
                polygonCount[str(i.end)] = [-1,None]
                                    
        newPolygonal = []
        for i in self.polygonal:
            if polygonCount[str(i)][0] == 3:
                for j in self.canonical:
                    if j.start == i:
                        lineStart = j
                    if j.end == i:
                        lineEnd = j
                line = bSegment(lineEnd.start, lineStart.end)
                self.canonical.remove(lineStart)
                self.canonical.remove(lineEnd)
                self.canonical.append(line)
            else:
                newPolygonal.append(i)
        self.canonical.sort()
        self.polygonal = newPolygonal
                        
                    
        
    def __iter__(self):
        return iter(self.canonical)
    
    def transform(self, matrix):
        result = boundary()
        for i in self.canonical:
            result.canonical.append(i.transform(matrix))
        for i in self.polygonal:
            result.polygonal.append(i.transform(matrix))
        return result
    
    def findFloatValue(self, other):
        for i in self.canonical:
            check = False
            for j in other.canonical:
                if i.start.vectorFloat == j.start and i.end.vectorFloat == j.end:
                    check = True
            if check == False:
                return False
        return True
        