# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 14:23:09 2016

@author: akibdpt
"""
from .individual import individual
from copy import deepcopy
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.vectorFloat import vector2Dfloat, vector3Dfloat
"""
 * A point is a 0-dimensional geometric data entity with a discrete
 * behavior.
 * 
 * The Point class extends on the Individual class.
 * It defines the characteristic individual for points.
 * A point is represented as a position vector.
 * This characteristic individual accepts no parameters.
 * It specifies a {sortal.map.similarityMap} as default.
 * Forms of points adhere to a discrete behavior.
"""
class point2D(individual):
    """
    initialize a point
    point contains: type: 'ind-point2D' and a coordinate vector e.g. (3,2)
    a point can either use vector2D or vector2Dfloat as it's coordinating vector
    vector2D: store numbers as symbolic expression e.g. sqrt(3), sin(pi/20), ...
    vector2Dfloat: store numbers as float values
    """
    def __init__(self, sort, x, y = None):
        super().__init__(sort)
        self.type = 'ind-point2D'
        if isinstance(x, vector2D) or isinstance(x, vector2Dfloat):
            self.value = x
        else:
            vector = vector2D(x,y)
            self.value = vector
            
    """
    print out a string representing the value of this point,
    used in form.printE()
    """
    def printVal(self):
        return 'x: '+str(self.value.x)+', y: '+str(self.value.y)   
         
    """
    check if the point's is (0,0)
    """
    def isNil(self):
        return self.value.isZero()
        
    """
    check equality
    """
    def __eq__(self, other):
        if other is None or other.type != 'ind-point2D' :
            return False
        if self.sort == other.sort:
            return (self.value == other.value)
        
    """
    comparing value of 2 points
    """
    def compareValue(self, other):
        return self.value.compareValue(other.value)   
        
    """
    duplicating a point
    """
    def duplicateInd(self):
        duplicate = point2D(self.sort, self.value)
        if self.attribute is not None:
            duplicate.addAttribute(self.attribute)
        return duplicate    
        
    """
    transform a point with a transform set
    """
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        return point2D(self.sort, self.value.transform(affine))
                
    """
    change the vector used from vector2D to vector2Dfloat
    """
    def convertFloat(self):
        point = point2D(self.sort, self.value.convertFloat())
        if self.hasAttribute():
            point.addAttribute(self.attribute.convertFloat())
        return point
    
    """
    check if the point has the same value as element after converting the
    individual into float
    """
    def findFloatValue(self, element):
        if element.type != "ind-point2D":
            return False
        if self.value.convertFloat() == element.value:
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    """
    check if the point belongs to a certain form after transformation with
    a transform set
    """
    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform2D)
        if affine is None:
            tfPoint = self
        else: 
            tfPoint =  point2D(self.sort, self.value.transform(affine))
            if self.hasAttribute():
                tfPoint.addAttribute(self.attribute)
        if compareForm.containsInd(tfPoint):
            return True
        return False



class point3D(individual):
    """
    initialize a point
    point contains: type: 'ind-point3D' and a coordinate vector e.g. (3,2)
    a point can either use vector3D or vector3Dfloat as it's coordinating vector
    vector3D: store numbers as symbolic expression e.g. sqrt(3), sin(pi/20), ...
    vector3Dfloat: store numbers as float values
    """
    def __init__(self, sort, x, y = None, z = None):
        super().__init__(sort)
        self.type = 'ind-point3D'
        if isinstance(x, vector3D) or isinstance(x, vector3Dfloat):
            self.value = x
        else:
            vector = vector3D(x,y,z)
            self.value = vector
            
    """
    print out a string representing the value of this point,
    used in form.printE()
    """
    def printVal(self):
        return str(self.value)
         
    """
    check if the point's is (0,0)
    """
    def isNil(self):
        return self.value.isZero()
        
    """
    check equality
    """
    def __eq__(self, other):
        if other is None or other.type != 'ind-point3D' :
            return False
        if self.sort == other.sort:
            return (self.value == other.value)
        
    """
    comparing value of 2 points
    """
    def compareValue(self, other):
        return self.value.compareValue(other.value)   
        
    """
    duplicating a point
    """
    def duplicateInd(self):
        duplicate = point3D(self.sort, self.value)
        if self.attribute is not None:
            for i in self.attribute:
                duplicate.addAttribute(i)
        return duplicate
        
    """
    transform a point with a transform set
    """
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        return point3D(self.sort, self.value.transform(affine))
                
    """
    change the vector used from vector3D to vector3Dfloat
    """
    def convertFloat(self):
        point = point3D(self.sort, self.value.convertFloat())
        if self.hasAttribute():
            for i in self.attribute:
                point.addAttribute(i.convertFloat())
        return point
    
    """
    check if the point has the same value as element after converting the
    individual into float
    """
    def findFloatValue(self, element):
        if element.type != "ind-point3D":
            return False
        if self.value.convertFloat() == element.value:
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    """
    check if the point belongs to a certain form after transformation with
    a transform set
    """
    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform3D)
        if affine is None:
            tfPoint = self.duplicateInd()
        else: 
            tfPoint =  point3D(self.sort, self.value.transform(affine))
            if self.attribute is not None:
                tfPoint.addAttribute(self.attribute)
        if compareForm.containsInd(tfPoint):
            return True
        return False
