# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 12:37:59 2016

@author: akibdpt
"""
from .mapping import mapping
from .distinguishable import distinguishable
from ..struct.vectorFloat import vector2Dfloat
from ..struct.coordinate import coordinate
from ..transform.transformSet import transformSet
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..ind.plane import plane3D
from ..ind.line import line2D, line3D
from ..ind.point import point2D, point3D
from enum import Enum
from copy import copy
from mpmath import mp

p = vector2Dfloat.p

class similarityMap2D(mapping):
    
    type = 'map-distinguishable-similarityMap2D'
    
        
    class casus(Enum):
        CASE_1A_2PT_NOT_COINCIDENT = 1,
        CASE_1C_1PT_1LN_NOT_COLINEAR = 2, """***"""
        CASE_1D_2LN_PARALLEL = 3,
        CASE_1FA_3LN_NO_TWO_PARALLEL = 4, 
        CASE_1FB_3LN_FIRST_TWO_PARALLEL = 5, 
        CASE_1FC_3LN_LAST_TWO_PARALLEL = 6,
        REDUNDANT = 0
        
    class distinguishes(Enum):
        INSUFFICIENT = 1, 
        UNSUITABLE = 2, 
        SUCCESS = 3

    def __init__(self):
        self.distinguishables = []
        self.transforms = []
        self.fromSet = [None,None,None]
        self.checkTree = None
        self.distinguishableSuccess = False
        self.reflection = 0
        self.transformset = transformSet()
        self.casus = similarityMap2D.casus.REDUNDANT
        self.subCase = []
        self.cosAngle = None
        self.ratio = None
        return
        
    @staticmethod
    def getInstance():
        return similarityMap2D()
    
    def clear(self):
        for i in self.distinguishables:
            i[1].clear()
        self.transforms = []
        return
    
    def map(self, fromInd, toInd):
        for i in self.distinguishables:
            if i[0] == fromInd:
                i[1].addCounterpart(toInd)
                return True
        return False
    
    def distinguish(self, ind):
        element = ind.codescriptor()
        dist = None
        for i in self.distinguishables:
            if element == i[0]:
                dist = i[1]
                dist.addInstance(ind)
                return dist
        if self.distinguishableSuccess:
            return None
        dist = distinguishable(element)
        
        cell =  [element, dist]    
        self.distinguishables.append(cell)
        dist.addInstance(ind)
        size = len(self.distinguishables)
        last = element
        first = None
        second = None
        #third = None

        if size > 1:
            first = self.distinguishables[0][0]
            if size > 2:
                second = self.distinguishables[1][0]
                #if size > 3:
                    #third = next(iterDistinguishables)
        success = self.assessDistinguishables(size, last, first, second)#, third)
        if success == similarityMap2D.distinguishes.UNSUITABLE:
            del self.distinguishables[-1]
        if success == similarityMap2D.distinguishes.SUCCESS:
            self.distinguishableSuccess = True
        return dist

    def assessDistinguishables(self, size, last, first, second):#, second, third):
        if size == 1:
            if last.__class__ == point2D or last.__class__ ==line2D:
                return similarityMap2D.distinguishes.INSUFFICIENT
        if size == 2 : #and first.__class__ == point2D:
            if first.__class__ == point2D and last.__class__ == line2D:
                self.fromSet[1] = first.value
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                if u.parallel(v):
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[0] = last.root.add(u.scale(v))     
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                self.casus = similarityMap2D.casus.CASE_1C_1PT_1LN_NOT_COLINEAR
                return similarityMap2D.distinguishes.SUCCESS         
                
            elif last.__class__ == line2D and first.__class__ == line2D:
                return similarityMap2D.distinguishes.INSUFFICIENT

            elif last.__class__ == point2D and first.__class__ == point2D:
                if last == first:
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[1] = last.value
                self.fromSet[0] = first.value                
                self.casus = similarityMap2D.casus.CASE_1A_2PT_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
        if size == 3:
            if last.__class__ == line2D and first.__class__ == line2D and second.__class__ == line2D:
                if not (last.parallel(first) or second.parallel(first) or second.parallel(last)):
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = first.intersection(second)
                    self.fromSet[2] = second.intersection(last)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[2]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))

                    u = self.fromSet[1].subtract(self.fromSet[0])
                    v = self.fromSet[2].subtract(self.fromSet[1])
                    w = self.fromSet[2].subtract(self.fromSet[0])
                    self.ratio = [u.squareLength(),v.squareLength(),w.squareLength()]
                    
                    self.checkTree = True                    
                    
                    self.casus = similarityMap2D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
                elif last.parallel(first) and second.parallel(first) and second.parallel(last):
                    return similarityMap2D.distinguishes.UNSUITABLE
                elif second.parallel(first):
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = second.intersection(last)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    
                    u = first.unitDir
                    v = last.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    self.checkTree = True                  
                    
                    self.casus = similarityMap2D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
                elif second.parallel(last):
                    self.fromSet[0] = second.intersection(first)
                    self.fromSet[1] = last.intersection(first)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    
                    u = first.unitDir
                    v = second.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    self.checkTree = True                    
                    
                    self.casus = similarityMap2D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
        return similarityMap2D.distinguishes.UNSUITABLE

    def assessTargetSet(self, target):
        if self.casus == similarityMap2D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        if self.casus == similarityMap2D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if not first.parallel(second):
                return False
            if second.compareValue(first) == 0:
                return False
            return True
        elif self.casus == similarityMap2D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            intersectingPoint = second.intersection(first)
            if not(second.containsPoint(intersectingPoint) and first.containsPoint(intersectingPoint)):
                return False
            return True
        return True
        
    def potentialTransforms(self): ########################
        if len(self.transforms) != 0:
            return self.transforms
        size = len(self.distinguishables)
        source = []
        counterparts = [] 
                        
        for element in self.distinguishables:
            source.append(element[1].element)
            counterparts.append(element[1].getCounterparts())
        target = [None]*size

        if self.checkTree and size == 2:
            self.checkTree = [[True for i in range(len(counterparts[1]))] 
                                    for j in range(len(counterparts[0]))] 
                                     
        elif self.checkTree and size == 3:
            self.checkTree = [[[True for i in range(len(counterparts[2]))] 
                                     for j in range(len(counterparts[1]))] 
                                     for k in range(len(counterparts[0]))]

        for e0 in range(len(counterparts[0])):
            target[0] = counterparts[0][e0]
            if size > 1:
                for e1 in range(len(counterparts[1])):
                    if counterparts[1][e1].individual == target[0].individual:
                        continue
                    target[1] = counterparts[1][e1]
                    if not self.assessTargetSet(target):
                        continue
                    if size > 2:
                        for e2 in range(len(counterparts[2])):
                            if self.checkTree:
                                if self.checkTree[e0][e1][e2]:
                                    self.checkTree[e0][e1][e2] = False
                                    self.checkTree[e0][e2][e1] = False
                                    self.checkTree[e1][e0][e2] = False
                                    self.checkTree[e1][e2][e0] = False
                                    self.checkTree[e2][e0][e1] = False
                                    self.checkTree[e2][e1][e0] = False
                                    target[2] = counterparts[2][e2]
                                    if (counterparts[2][e2].individual == target[1].individual or 
                                        counterparts[2][e2].individual == target[0].individual):
                                        continue
                                    self.detectTransforms(source, target)
                            else:
                                target[2] = counterparts[2][e2]
                                self.detectTransforms(source, target)
                    else:
                        if self.checkTree:
                            if self.checkTree[e0][e1]:
                                self.checkTree[e0][e1] = False
                                self.checkTree[e1][e0] = False
                                self.detectTransforms(source, target)
                        else:
                            self.detectTransforms(source, target)
        return self.transforms
        
    def detectTransforms(self, source, target):
        size = len(source)
        #leftLast = source[-1].codescriptor()
        last = target[-1].individual
        if size > 1:
            #leftFirst = source[0].codescriptor()
            first = target[0].individual
            if size > 2:
                second = target[1].individual
        #type = last.__class__
        toSet = [None, None, None]
            #########################################################
        if self.casus == similarityMap2D.casus.CASE_1C_1PT_1LN_NOT_COLINEAR:
            if last.type == 'ind-point2D':
                temp = first
                first = last
                last = temp
            toSet[1] = first.value
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return False                
            toSet[0] = last.root.add(u.scale(v))
            
            if round(toSet[0].subtract(toSet[1]).squareLength(),12)==0:
                return False
                
                #self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                #self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                
            if (last.assoc.containsPoint(toSet[0]) >= self.subCase[0] and
                (not(last.assoc.start == toSet[0] or last.assoc.end == toSet[0])) >= self.subCase[1]):

                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), copy(target), tf.index])
                        """
                        transf = [tf, copy(source), copy(target), tf.index]
                        if transf not in self.transforms:
                            self.transforms.append(transf)
                        """
            return True

        
            #########################################################
        elif self.casus == similarityMap2D.casus.CASE_1D_2LN_PARALLEL:
            if not last.parallel(first):
                return False
            toSet[0] = last.assoc.start
            u = first.direction
            v = toSet[0].subtract(first.root)
            if u.parallel(v):
                return similarityMap2D.distinguishes.UNSUITABLE
            toSet[1] = first.root.add(u.scale(v))
            
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append([tf, copy(source), copy(target), tf.index])
            
            toSet[0] = last.assoc.end
            u = first.direction
            v = toSet[0].subtract(first.root)
            if u.parallel(v):
                return similarityMap2D.distinguishes.UNSUITABLE
            toSet[1] = first.root.add(u.scale(v))
            
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append([tf, copy(source), copy(target), tf.index+2])
            return True
            #########################################################
            
        elif self.casus == similarityMap2D.casus.CASE_1A_2PT_NOT_COINCIDENT:
            if last == first:
                return False
            toSet[1] = last.value
            toSet[0] = first.value
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append([tf, copy(source), copy(target), tf.index])
            return True
            #########################################################
            
        elif self.casus == similarityMap2D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL:
            if (last.parallel(first) or second.parallel(first) or second.parallel(last)):
                return False
            is1 = first.intersection(last)
            is2 = first.intersection(second)
            is3 = second.intersection(last)
            if is1 == is2 or is1 == is3 or is2 == is3:
                return False
            u = is2.subtract(is1).squareLength()
            v = is3.subtract(is2).squareLength()
            w = is3.subtract(is1).squareLength()
            u0= self.ratio[0]
            v0= self.ratio[1]
            w0= self.ratio[2]
            ts = []

            if (round(u/u0,p) == round(v/v0,p) and round(u/u0,p) == round(w/w0,p)):
                ts.append([is1,is2,is3,target[0],target[1],target[2]])
            if (round(u/u0,p) == round(w/v0,p) and round(u/u0,p) == round(v/w0,p)):
                ts.append([is2,is1,is3,target[0],target[2],target[1]])
            if (round(v/u0,p) == round(u/v0,p) and round(v/u0,p) == round(w/w0,p)):
                ts.append([is3,is2,is1,target[1],target[0],target[2]])
            if (round(v/u0,p) == round(w/v0,p) and round(v/u0,p) == round(u/w0,p)):
                ts.append([is2,is3,is1,target[1],target[2],target[0]])
            if (round(w/u0,p) == round(u/v0,p) and round(w/u0,p) == round(v/w0,p)):
                ts.append([is3,is1,is2,target[2],target[0],target[1]])
            if (round(w/u0,p) == round(v/v0,p) and round(w/u0,p) == round(u/w0,p)):
                ts.append([is1,is3,is2,target[2],target[1],target[0]])
                
            for toSet in ts:
                for i in toSet[3].forms[toSet[3].individual.sort.name].elements:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                for i in toSet[4].forms[toSet[3].individual.sort.name].elements:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                for i in toSet[5].forms[toSet[3].individual.sort.name].elements:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    continue

                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), [toSet[3],toSet[4],toSet[5]], tf.index])
                        
            #########################################################
        elif self.casus == similarityMap2D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL:
            if (last.parallel(first) and second.parallel(first) and second.parallel(last)):
                return False
            if not second.parallel(first):
                if second.parallel(last):
                    temp = last
                    last = first
                    first = temp
                    temp = target[2]
                    target[2] = target[0]
                    target[0] = temp
                else:
                    return False
                
            u = first.unitDir
            v = last.unitDir
            cosA = abs(u.dotproduct(v))
            if round(self.cosAngle,p) !=  round(cosA,p):
                return False
            
            toSet[0] = first.intersection(last)
            toSet[1] = second.intersection(last)
            
            ###########
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[0]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[1]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[2] and
                    i.containsPoint(toSet[1]) >= self.subCase[3]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:               
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), copy(target), tf.index])       
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[1]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[0]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[3] and
                    i.containsPoint(toSet[1]) >= self.subCase[2]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:               
                tfs = affineTransform2D.similarityTransform(self.fromSet, [toSet[1],toSet[0],None], False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), [target[1],target[0],target[2]], tf.index])    
            #------------------------------------------------------------    
            #########################################################
        elif self.casus == similarityMap2D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL:
            if (last.parallel(first) and second.parallel(first) and second.parallel(last)):
                return False
            if not second.parallel(last):
                if second.parallel(first):
                    temp = last
                    last = first
                    first = temp
                    temp = target[2]
                    target[2] = target[0]
                    target[0] = temp
                else:
                    return False
                
            u = first.unitDir
            v = second.unitDir
            cosA = abs(u.dotproduct(v))
            if round(self.cosAngle,p) !=  round(cosA,p):
                return False
            
            toSet[0] = second.intersection(first)
            toSet[1] = last.intersection(first)
            
            ###########
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    i.containsPoint(toSet[1]) >= self.subCase[1]):
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[2]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[3]:
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:               
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), copy(target), tf.index])       
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[1] and
                    i.containsPoint(toSet[1]) >= self.subCase[0]):
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[3]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[2]:
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:               
                tfs = affineTransform2D.similarityTransform(self.fromSet, [toSet[1],toSet[0],None], False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), [target[0],target[2],target[1]], tf.index])    
            """
            if (last.parallel(first) and second.parallel(first) and second.parallel(last)):
                return False
            if not second.parallel(last):
                return False
                
            toSet[0] = second.intersection(first)
            toSet[1] = last.intersection(first)
            
            if     (first.assoc.containsPoint(toSet[0]) < self.subCase[0] or
                    first.assoc.containsPoint(toSet[1]) < self.subCase[1] or
                    second.assoc.containsPoint(toSet[0]) < self.subCase[2] or
                    last.assoc.containsPoint(toSet[1]) < self.subCase[3]):
                return False
            
            u = first.unitDir
            v = second.unitDir
            cosA = abs(u.dotproduct(v))
            if round(self.cosAngle,p) !=  round(cosA,p):
                return False
            
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, False, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append([tf, copy(source), copy(target), tf.index])
            """
        return False
        
    def replicateTransform(self, source, target, index):
        size = len(source)
        leftLast = source[-1].codescriptor()
        last = target[-1].codescriptor()
        if size > 1:
            leftFirst = source[0].codescriptor()
            first = target[0].codescriptor()
            if size > 2:
                leftSecond = source[1].codescriptor()
                second = target[1].codescriptor()
        #type = last.__class__
        fromSet = [None, None, None]
        toSet = [None, None, None]
        transform = []
        if self.casus == similarityMap2D.casus.CASE_1C_1PT_1LN_NOT_COLINEAR:
                
            fromSet[1] = leftFirst.value.simplify()
            u = leftLast.direction
            v = fromSet[1].subtract(leftLast.root)
            fromSet[0] = leftLast.root.add(u.scale(v)).simplify()
            
            toSet[1] = first.value.simplify()
            u = last.direction
            v = toSet[1].subtract(last.root)
            toSet[0] = last.root.add(u.scale(v)).simplify()
            
            tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
            return transform


            #########################################################
        elif self.casus == similarityMap2D.casus.CASE_1D_2LN_PARALLEL:
            fromSet[0] = leftLast.assoc.start.simplify()
            u = leftFirst.direction
            v = fromSet[0].subtract(leftFirst.root)
            fromSet[1] = leftFirst.root.add(u.scale(v)).simplify()
                
            if index <= 2:   
                toSet[0] = last.assoc.start.simplify()
                u = first.direction
                v = toSet[0].subtract(first.root)
                toSet[1] = first.root.add(u.scale(v)).simplify()
                
                tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index)
                for tf in tfs:
                    if tf is not None:
                        transform.append(tf)
            else:
                toSet[0] = last.assoc.end.simplify()
                u = first.direction
                v = toSet[0].subtract(first.root)
                toSet[1] = first.root.add(u.scale(v)).simplify()
                
                tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index-2)
                for tf in tfs:
                    if tf is not None:
                        transform.append(tf)
            return transform

        elif self.casus == similarityMap2D.casus.CASE_1A_2PT_NOT_COINCIDENT:
            fromSet[1] = leftLast.value.simplify()
            fromSet[0] = leftFirst.value.simplify()
            
            toSet[1] = last.value.simplify()
            toSet[0] = first.value.simplify()
            
            tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
            return transform
        elif self.casus == similarityMap2D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL:
            fromSet[0] = leftFirst.intersection(leftLast).simplify()
            fromSet[1] = leftFirst.intersection(leftSecond).simplify()
            fromSet[2] = leftSecond.intersection(leftLast).simplify()
            
            toSet[0] = first.intersection(last).simplify()
            toSet[1] = first.intersection(second).simplify()
            toSet[2] = second.intersection(last).simplify()
            tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
                    
        elif self.casus == similarityMap2D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL:
            fromSet[0] = leftFirst.intersection(leftLast).simplify()
            fromSet[1] = leftSecond.intersection(leftLast).simplify()
            toSet[0] = first.intersection(last).simplify()
            toSet[1] = second.intersection(last).simplify()
            tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
                    
        elif self.casus == similarityMap2D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL:
            fromSet[0] = leftSecond.intersection(leftFirst).simplify()
            fromSet[1] = leftLast.intersection(leftFirst).simplify()         
            toSet[0] = last.intersection(first).simplify()
            toSet[1] = second.intersection(first).simplify()
            tfs = affineTransform2D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
                    
        return transform
        
    def checkTransform(self):
        return
     
     
     
     

    
class similarityMap3D(mapping):
    
    type = 'map-distinguishable-similarityMap3D'

    class casus(Enum):
        CASE_1A_3PT_NOT_COLINEAR = 1,  #3
        CASE_1B_2PT_1PL_NOT_PERPENDICULAR = 2, #3 ***
        CASE_1C_1PT_1LN_NOT_COLINEAR = 3, #2
        CASE_1D_1PT_1LN_1PL_NOT_COINCIDENT_NOT_PERPENDICULAR = 4,  #3 ***
        CASE_1E_1PT_2PL_NOT_CONCURRENT = 5, #3 ***
        CASE_1FA_3LN_NO_TWO_PARALLEL = 6,  #3
        CASE_1FB_3LN_FIRST_TWO_PARALLEL = 7,  #3
        CASE_1FC_3LN_LAST_TWO_PARALLEL = 8, #3
        CASE_1G_2LN_SKEWED = 9, #2 ***
        CASE_1H_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT = 10, #3 ***
        CASE_1I_2LN_1PL_NOT_PARALLEL_NOR_COINCIDENT = 11, #3 ***
        CASE_1J_1LN_2PL_NOT_ALL_CONCURRENT = 12, #3 ***
        CASE_1K_4PL_NOT_ALL_PARALLEL = 13,  #4 ***
        REDUNDANT = 0

    class distinguishes(Enum):
        INSUFFICIENT = 1, 
        UNSUITABLE = 2, 
        SUCCESS = 3

    def __init__(self):
        self.distinguishables = []
        self.transforms = []
        self.fromSet = [None,None,None]
        self.checkTree = None
        self.distinguishableSuccess = False
        self.reflection = 0
        self.transformset = transformSet()
        self.casus = similarityMap3D.casus.REDUNDANT
        self.subCase = []
        self.cosAngle = None
        self.ratio = None
        return
        
    @staticmethod
    def getInstance():
        return similarityMap3D()
    
    def clear(self):
        for i in self.distinguishables:
            i[1].clear()
        self.transforms = []
        return
    
    def map(self, fromInd, toInd):
        for i in self.distinguishables:
            if i[0] == fromInd:
                i[1].addCounterpart(toInd)
                return True
        return False
    
    def distinguish(self, ind):
        element = ind.codescriptor()
        dist = None
        for i in self.distinguishables:
            if element == i[0]:
                dist = i[1]
                dist.addInstance(ind)
                return dist
        if self.distinguishableSuccess:
            return None
        dist = distinguishable(element)
        
        cell =  [element, dist]    
        self.distinguishables.append(cell)
        dist.addInstance(ind)
        size = len(self.distinguishables)
        last = element
        first = None
        second = None
        third = None
        if size > 1:
            first = self.distinguishables[0][0]
            if size > 2:
                second = self.distinguishables[1][0]
                if size > 3:
                    third = self.distinguishables[2][0]
        success = self.assessDistinguishables(size, last, first, second, third)
        
        if success == similarityMap3D.distinguishes.UNSUITABLE:
            del self.distinguishables[-1]
        if success == similarityMap3D.distinguishes.SUCCESS:
            self.distinguishableSuccess = True
        return dist

    def assessDistinguishables(self, size, last, first, second, third):
        if size == 1:
            if last.__class__ == point3D or last.__class__ ==line3D:
                return similarityMap3D.distinguishes.INSUFFICIENT
        if size == 2 : #and first.__class__ == point3D:
            if first.__class__ == point3D and last.__class__ == line3D:
                
                self.fromSet[1] = first.value
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                if u.parallel(v):
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[0] = last.root.add(u.scale(v))     
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                d = self.fromSet[1].subtract(self.fromSet[0])
                self.fromSet[2] = self.fromSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**mp.mpf('0.5'))).simplify()

                if not last.assoc.containsPoint(self.fromSet[2]):
                    self.fromSet[2] = self.fromSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**mp.mpf('0.5'))).simplify()
                
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                
                self.casus = similarityMap3D.casus.CASE_1C_1PT_1LN_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS         
                
            elif last.__class__ == line3D and first.__class__ == line3D:
                return similarityMap3D.distinguishes.INSUFFICIENT


            elif last.__class__ == point3D and first.__class__ == point3D:
                if last == first:
                    return similarityMap3D.distinguishes.UNSUITABLE
                return similarityMap3D.distinguishes.INSUFFICIENT
                
        if size == 3:
            if last.__class__ == point3D and first.__class__ == point3D and second.__class__ == point3D:
                if last == first or last == second:
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[0] = first.value      
                self.fromSet[1] = second.value                
                self.fromSet[2] = last.value
                ################### need subcase evaluation ###################
                self.casus = similarityMap2D.casus.CASE_1A_2PT_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
            
            if last.__class__ == plane3D and first.__class__ == point3D and second.__class__ == point3D:
                foot = last.projection(first)
                if foot == first:
                    foot = last.projection(second)
                    if foot == second:
                        d = second.subtract(first)
                        u = last.normal.product(d)
                        foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u))) ** mp.mpf('0.5')))
                        self.subCase = 0
                    else:
                        self.subCase = 2
                else:
                    self.subCase = 1

                if (foot.colinear(second, first)):
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[0] = first.value
                self.fromSet[1] = second.value
                self.fromSet[2] = foot.value
                self.casus = similarityMap3D.casus.CASE_1B_2PT_1PL_NOT_PERPENDICULAR
                return similarityMap3D.distinguishes.SUCCESS
            
            if last.__class__ == line3D and first.__class__ == line3D and second.__class__ == line3D:
                if not (last.parallel(first) or second.parallel(first) or second.parallel(last)):

                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = first.intersection(second)
                    self.fromSet[2] = second.intersection(last)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[2]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))

                    u = self.fromSet[1].subtract(self.fromSet[0])
                    v = self.fromSet[2].subtract(self.fromSet[1])
                    w = self.fromSet[2].subtract(self.fromSet[0])
                    self.ratio = [u.squareLength(),v.squareLength(),w.squareLength()]
                    
                    self.checkTree = True                    
                    
                    self.casus = similarityMap3D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS


                elif last.parallel(first) and second.parallel(first) and second.parallel(last):
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                elif second.parallel(first):
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = second.intersection(last)
                    
                    u = first.direction
                    d = self.fromSet[1].subtract(self.fromSet[0])
                    v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** mp.mpf('0.5'))
                    self.fromSet[2] = self.fromSet[0].add(v)
                    if not first.assoc.containsPoint(self.fromSet[2]):
                        self.fromSet[2] = self.fromSet[0].subtract(v)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[2]))
                    
                    u = first.unitDir
                    v = last.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    self.checkTree = True       
                
                    self.casus = similarityMap3D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS
                elif second.parallel(last):
                    self.fromSet[0] = second.intersection(first)
                    self.fromSet[1] = last.intersection(first)
                    
                    u = second.direction
                    d = self.fromSet[0].subtract(self.fromSet[1])
                    v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** mp.mpf('0.5'))
                    self.fromSet[2] = self.fromSet[1].add(v)
                    if not last.assoc.containsPoint(self.fromSet[2]):
                        self.fromSet[2] = self.fromSet[1].subtract(v)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                    
                    u = first.unitDir
                    v = second.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    self.checkTree = True                    
                    
                    self.casus = similarityMap3D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS
                    
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                    
        #raise NotImplementedError('not yet')
            ####################################
                    
        return similarityMap3D.distinguishes.UNSUITABLE

    def assessTargetSet(self, target):
        if self.casus == similarityMap3D.casus.CASE_1A_3PT_NOT_COLINEAR:
            first = target[0].individual
            second = target[1].individual
            if first == second:
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if not first.parallel(second):
                return False
            if second.compareValue(first) == 0:
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            intersectingPoint = second.intersection(first)
            if not(second.containsPoint(intersectingPoint) and first.containsPoint(intersectingPoint)):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_1B_2PT_1PL_NOT_PERPENDICULAR:
            first = target[0].individual
            second = target[1].individual
            if first.type == second.type and first == second:
                return False
        return True
        
    def potentialTransforms(self): ########################
        if len(self.transforms) != 0:
            return self.transforms
        size = len(self.distinguishables)
        source = []
        counterparts = [] 
                        
        for element in self.distinguishables:
            source.append(element[1].element)
            counterparts.append(element[1].getCounterparts())
        target = [None]*size

        if self.checkTree and size == 2:
            self.checkTree = [[True for i in range(len(counterparts[1]))] 
                                    for j in range(len(counterparts[0]))] 
                                     
        elif self.checkTree and size == 3:
            self.checkTree = [[[True for i in range(len(counterparts[2]))] 
                                     for j in range(len(counterparts[1]))] 
                                     for k in range(len(counterparts[0]))]
                       
        """               
        print()
        for e0 in range(len(counterparts)):
            print('\ncounterpart set '+str(e0)+':')
            for e1 in range(len(counterparts[e0])):
                print(counterparts[e0][e1].individual)
        """

        for e0 in range(len(counterparts[0])):
            target[0] = counterparts[0][e0]
            if size > 1:
                for e1 in range(len(counterparts[1])):
                    if counterparts[1][e1].individual == target[0].individual:
                        continue
                    target[1] = counterparts[1][e1]
                    if not self.assessTargetSet(target):
                        continue
                    if size > 2:
                        for e2 in range(len(counterparts[2])):
                            if self.checkTree:
                                if self.checkTree[e0][e1][e2]:
                                    self.checkTree[e0][e1][e2] = False
                                    self.checkTree[e0][e2][e1] = False
                                    self.checkTree[e1][e0][e2] = False
                                    self.checkTree[e1][e2][e0] = False
                                    self.checkTree[e2][e0][e1] = False
                                    self.checkTree[e2][e1][e0] = False
                                    target[2] = counterparts[2][e2]
                                    if (counterparts[2][e2].individual == target[1].individual or 
                                        counterparts[2][e2].individual == target[0].individual):
                                        continue
                                    self.detectTransforms(source, target)
                            else:
                                target[2] = counterparts[2][e2]
                                self.detectTransforms(source, target)
                    else:
                        if self.checkTree:
                            if self.checkTree[e0][e1]:
                                self.checkTree[e0][e1] = False
                                self.checkTree[e1][e0] = False
                                self.detectTransforms(source, target)
                        else:
                            self.detectTransforms(source, target)
        return self.transforms
        
    def detectTransforms(self, source, target):

        size = len(source)
        #leftLast = source[-1].codescriptor()
        last = target[-1].individual
        if size > 1:
            #leftFirst = source[0].codescriptor()
            first = target[0].individual
            if size > 2:
                second = target[1].individual
        #type = last.__class__
        toSet = [None, None, None]
            #########################################################

        if self.casus == similarityMap3D.casus.CASE_1C_1PT_1LN_NOT_COLINEAR:
            if last.type == 'ind-point3D':
                temp = first
                first = last
                last = temp
            toSet[1] = first.value
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return False                
            toSet[0] = last.root.add(u.scale(v))
            
            if round(toSet[0].subtract(toSet[1]).squareLength(),12)==0:
                return False
            
            ####################### need revision
            for i in target[-1].forms[target[-1].individual.sort.name].elements:
                check1 = False
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check1 = True
                    break
            if check1:
                d = toSet[1].subtract(toSet[0])
                
                toSet[2] = toSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**mp.mpf('0.5'))).simplify()
                
                for i in target[-1].forms[target[-1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), copy(target), tf.index])
                        break
    
                        
                toSet[2] = toSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**mp.mpf('0.5'))).simplify()
                for i in target[-1].forms[target[-1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), copy(target), tf.index+2])
                        break

        elif self.casus == similarityMap3D.casus.CASE_1A_3PT_NOT_COLINEAR:
            if last == first or last == second or first == second:
                return False
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = last.value
            ##################### need subcase evaluation #####################
                    
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append([tf, copy(source), copy(target), tf.index])
            return True
            #########################################################
            
        elif self.casus == similarityMap3D.casus.CASE_1B_2PT_1PL_NOT_PERPENDICULAR:
            foot = last.projection(first)
            if foot == first: 
                if self.subCase == 1:
                    return False
                else:
                    foot = last.projection(second)
                    if foot == second:
                        if self.subCase == 2:
                            return False
                        else:
                            d = second.subtract(first)
                            u = last.normal.product(d)
                            foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u))) ** mp.mpf('0.5')))

            if (foot.colinear(second, first)):
                return False
                
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = foot.value
            
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append([tf, copy(source), copy(target), tf.index])
            return True
            
        elif self.casus == similarityMap3D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL:
            if (last.parallel(first) or second.parallel(first) or second.parallel(last)):
                return False
            
            is1 = first.intersection(last)
            is2 = first.intersection(second)
            is3 = second.intersection(last)
            
            if is1 == is2 or is1 == is3 or is2 == is3:
                return False
            u = is2.subtract(is1).squareLength()
            v = is3.subtract(is2).squareLength()
            w = is3.subtract(is1).squareLength()
            u0= self.ratio[0]
            v0= self.ratio[1]
            w0= self.ratio[2]
            ts = []

            if (round(u/u0,p) == round(v/v0,p) and round(u/u0,p) == round(w/w0,p)):
                ts.append([is1,is2,is3,target[0],target[1],target[2]])
            if (round(u/u0,p) == round(w/v0,p) and round(u/u0,p) == round(v/w0,p)):
                ts.append([is2,is1,is3,target[0],target[2],target[1]])
            if (round(v/u0,p) == round(u/v0,p) and round(v/u0,p) == round(w/w0,p)):
                ts.append([is3,is2,is1,target[1],target[0],target[2]])
            if (round(v/u0,p) == round(w/v0,p) and round(v/u0,p) == round(u/w0,p)):
                ts.append([is2,is3,is1,target[1],target[2],target[0]])
            if (round(w/u0,p) == round(u/v0,p) and round(w/u0,p) == round(v/w0,p)):
                ts.append([is3,is1,is2,target[2],target[0],target[1]])
            if (round(w/u0,p) == round(v/v0,p) and round(w/u0,p) == round(u/w0,p)):
                ts.append([is1,is3,is2,target[2],target[1],target[0]])
                
            for toSet in ts:
                for i in toSet[3].forms[toSet[3].individual.sort.name].elements:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                for i in toSet[4].forms[toSet[3].individual.sort.name].elements:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                for i in toSet[5].forms[toSet[3].individual.sort.name].elements:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    continue

                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append([tf, copy(source), [toSet[3],toSet[4],toSet[5]], tf.index])
                        
            #########################################################

        elif self.casus == similarityMap3D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL:
            if (last.parallel(first) and second.parallel(first) and second.parallel(last)):
                return False
            if not second.parallel(first):
                if second.parallel(last):
                    temp = last
                    last = first
                    first = temp
                    temp = target[2]
                    target[2] = target[0]
                    target[0] = temp
                else:
                    return False
                
            u = first.unitDir
            v = last.unitDir
            cosA = abs(u.dotproduct(v))
            if round(self.cosAngle,p) !=  round(cosA,p):
                return False
            
            toSet[0] = first.intersection(last)
            toSet[1] = second.intersection(last)
            
            ###########
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[0]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[1]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[2] and
                    i.containsPoint(toSet[1]) >= self.subCase[3]):
                    check3 = True
                    break
                check3 = False

            if check1 and check2 and check3:        
                
                
                u = first.direction
                d = toSet[1].subtract(toSet[0])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** mp.mpf('0.5'))
                toSet[2] = toSet[0].add(v)
                
                for i in target[0].forms[target[0].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), copy(target), tf.index])     
                        break
                    
                toSet[2] = toSet[0].subtract(v)
                for i in target[0].forms[target[0].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), copy(target), tf.index+10])   
                        break
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[1]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[0]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[3] and
                    i.containsPoint(toSet[1]) >= self.subCase[2]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:  
                
                u = second.direction
                d = toSet[0].subtract(toSet[1])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** mp.mpf('0.5'))
                toSet[2] = toSet[1].add(v)
                                
                for i in target[1].forms[target[1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, [toSet[1],toSet[0], toSet[2]], False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), [target[1],target[0],target[2]], tf.index])  
                
                toSet[2] = toSet[1].subtract(v)
                
                for i in target[1].forms[target[1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, [toSet[1],toSet[0], toSet[2]], False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), [target[1],target[0],target[2]], tf.index+10])
            #------------------------------------------------------------    
            #########################################################
            #########################################################
            ##################    Not done yet    ###################
            #########################################################
            #########################################################
        elif self.casus == similarityMap3D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL:
            if (last.parallel(first) and second.parallel(first) and second.parallel(last)):
                return False
            if not second.parallel(last):
                if second.parallel(first):
                    temp = last
                    last = first
                    first = temp
                    temp = target[2]
                    target[2] = target[0]
                    target[0] = temp
                else:
                    return False
                
            u = first.unitDir
            v = second.unitDir
            cosA = abs(u.dotproduct(v))
            if round(self.cosAngle,p) !=  round(cosA,p):
                return False
            
            toSet[0] = second.intersection(first)
            toSet[1] = last.intersection(first)
            
            ###########
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    i.containsPoint(toSet[1]) >= self.subCase[1]):
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[2]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[3]:
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:       
                
                u = second.direction
                d = toSet[0].subtract(toSet[1])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** mp.mpf('0.5'))
                toSet[2] = toSet[1].add(v)
                                
                for i in target[1].forms[target[1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), copy(target), tf.index])     
                
                toSet[2] = toSet[1].subtract(v)
                
                for i in target[1].forms[target[1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), copy(target), tf.index+10])     
                                        
                        
                        
                        
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name].elements:
                if (i.containsPoint(toSet[0]) >= self.subCase[1] and
                    i.containsPoint(toSet[1]) >= self.subCase[0]):
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name].elements:
                if i.containsPoint(toSet[0]) >= self.subCase[3]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name].elements:
                if i.containsPoint(toSet[1]) >= self.subCase[2]:
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:       
                
                u = last.direction
                d = toSet[1].subtract(toSet[0])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** mp.mpf('0.5'))
                toSet[2] = toSet[0].add(v)
                                
                for i in target[-1].forms[target[-1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, [toSet[1],toSet[0], toSet[2]], False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), [target[1],target[0],target[2]], tf.index])  
                
                toSet[2] = toSet[1].subtract(v)
                
                for i in target[-1].forms[target[-1].individual.sort.name].elements:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, [toSet[1],toSet[0], toSet[2]], False, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append([tf, copy(source), [target[1],target[0],target[2]], tf.index+10])
            #########################################################
            #########################################################
            #########################################################
            #########################################################
            #########################################################

        return False
        
    def replicateTransform(self, source, target, index):
        size = len(source)
        leftLast = source[-1].codescriptor()
        last = target[-1].codescriptor()
        if size > 1:
            leftFirst = source[0].codescriptor()
            first = target[0].codescriptor()
            if size > 2:
                leftSecond = source[1].codescriptor()
                second = target[1].codescriptor()
        #type = last.__class__
        fromSet = [None, None, None]
        toSet = [None, None, None]
        transform = []
        if self.casus == similarityMap3D.casus.CASE_1C_1PT_1LN_NOT_COLINEAR:
                
            fromSet[1] = leftFirst.value.simplify()
            u = leftLast.direction
            v = fromSet[1].subtract(leftLast.root)
            fromSet[0] = leftLast.root.add(u.scale(v)).simplify()
            d = fromSet[1].subtract(fromSet[0])
            fromSet[2] = fromSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**coordinate(0.5))).simplify()

            if not last.assoc.containsPoint(fromSet[2]):
                fromSet[2] = fromSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**coordinate(0.5))).simplify()
                
            toSet[1] = first.value.simplify()
            u = last.direction
            v = toSet[1].subtract(last.root)
            toSet[0] = last.root.add(u.scale(v)).simplify()


            d = toSet[1].subtract(toSet[0])
            if index <=2:
                toSet[2] = toSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**coordinate(0.5))).simplify()
            else:
                toSet[2] = toSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**coordinate(0.5))).simplify()
                index -= 2
            tfs = affineTransform3D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)

            return transform
        elif self.casus == similarityMap3D.casus.CASE_1B_2PT_1PL_NOT_PERPENDICULAR:
            if self.subCase == 1:
                leftFoot = leftLast.projection(leftFirst)
                foot = last.projection(first)
            elif self.subCase == 2:
                leftFoot = leftLast.projection(leftSecond)
                foot = last.projection(second)
            else:
                d = leftSecond.subtract(leftFirst)
                u = leftLast.normal.product(d)
                leftFoot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u))) ** coordinate(0.5)))
                d = second.subtract(first)
                u = last.normal.product(d)
                foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u))) ** coordinate(0.5)))

            fromSet[0] = leftFirst.value
            fromSet[1] = leftSecond.value
            fromSet[2] = leftFoot.value
            
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = foot.value
            
            tfs = affineTransform3D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)

        elif self.casus == similarityMap3D.casus.CASE_1FA_3LN_NO_TWO_PARALLEL:
            fromSet[0] = leftFirst.intersection(leftLast).simplify()
            fromSet[1] = leftFirst.intersection(leftSecond).simplify()
            fromSet[2] = leftSecond.intersection(leftLast).simplify()
            
            toSet[0] = first.intersection(last).simplify()
            toSet[1] = first.intersection(second).simplify()
            toSet[2] = second.intersection(last).simplify()
            tfs = affineTransform3D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
                    
        elif self.casus == similarityMap3D.casus.CASE_1FB_3LN_FIRST_TWO_PARALLEL:
            fromSet[0] = leftFirst.intersection(leftLast).simplify()
            fromSet[1] = leftSecond.intersection(leftLast).simplify()
            
            u = leftFirst.direction
            d = fromSet[1].subtract(fromSet[0])
            v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** coordinate(0.5))
            fromSet[2] = fromSet[0].add(v).simplify()
            if not first.assoc.containsPoint(fromSet[2]):
                fromSet[2] = fromSet[0].subtract(v).simplify()
            
            toSet[0] = first.intersection(last).simplify()
            toSet[1] = second.intersection(last).simplify()
            
            u = first.direction
            d = toSet[1].subtract(toSet[0])
            v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** coordinate(0.5))

            if index > 10:
                toSet[2] = toSet[0].subtract(v).simplify()
                index -= 10
            else:
                toSet[2] = toSet[0].add(v).simplify()
            tfs = affineTransform3D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)
                    
        elif self.casus == similarityMap3D.casus.CASE_1FC_3LN_LAST_TWO_PARALLEL:
            fromSet[0] = leftSecond.intersection(leftFirst).simplify()
            fromSet[1] = leftLast.intersection(leftFirst).simplify()        
            
            u = leftSecond.direction
            d = fromSet[0].subtract(fromSet[1])
            v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** coordinate(0.5))
            
            fromSet[2] = fromSet[1].add(v).simplify()
            if not first.assoc.containsPoint(fromSet[2]):
                fromSet[2] = fromSet[1].subtract(v).simplify()
            
            toSet[0] = last.intersection(first).simplify()
            toSet[1] = second.intersection(first).simplify()
            
            u = second.direction
            d = toSet[0].subtract(toSet[1])
            v = u.scale((d.dotproduct(d)/(u.dotproduct(u))) ** coordinate(0.5))
            if index > 10:
                toSet[2] = toSet[1].subtract(v).simplify()
                index -= 10
            else:
                toSet[2] = toSet[1].add(v).simplify()
            
            tfs = affineTransform3D.similarityTransform(fromSet, toSet, True, index)
            for tf in tfs:
                if tf is not None:
                    transform.append(tf)

        return transform
        