# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 09:22:12 2016

@author: akibdpt
"""
from .simpleSort import simpleSort

"""
 * An Attribute Sort specifies a composition of a simple {Sort}
 * (its base sort) with any other {Sort} (its weigth sort)
 * under the object-attribute relationship. An individual of the
 * attribute sort is an individual of the base sort that has a form of the
 * weight sort assigned as an attribute.
 * An attribute sort may have a name assigned.
 * 
 * The AttributeSort class represents an attribute sort additionally by
 * its base and weight sorts. The canonical version of an attribute sort is the
 * unnamed attribute sort of the canonical versions of the base and weight sort.
"""
class attributeSort(simpleSort):
    separator = ' ^ '
    
    """
    initialize the attribute sort with name, description, base sort and weight sort
    """
    def __init__(self, name, base, weight):
        if self.checkSort(name) == False:
            pass
            raise ValueError('sort name already exists')
        super().__init__(name)
        self.base = base
        self.weight = weight
        self.description = self.description(base, weight)
        self.type = 'sort-attribute'
        return

    """
    returning the base sort
    """
    def baseSort(self):
        return self.base

    """
    creating the description of the sort
    template: BaseSort ^ WeightSort
    e.g. point ^ label, line ^ weight, lineSegment ^ (label + weight + color)
    """
    def description(self, base, weight):
        desc = base.name + attributeSort.separator
        desc += weight.name
        return desc
        
    """
    check if this sort contains other sort
    check both basesort and weight sort if this sort is not the same as other sort
    """
    def contains(self, other):
        if other.name == self.name:
            return True
        if other.type == 'sort-attribute' and self.base.name == other.base.name:
            return self.weight.contains(other.weight)
        return False
        
    """
    creating a new Form (discrete form / ordinal form / interval form) of this sort
    """
    def newForm(self):
        try:
            form = self.base.behavior(self)
            return form
        except:
            print('exception')
    
    