# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 14:33:20 2016

@author: akibdpt
"""
from sympy import sympify, S, sqrt, simplify, expand, N
from decimal import Decimal, getcontext
from fractions import Fraction

"""
 * Coordinate is the representational unit for geometric
 * Coordinate is saved and represented as symbolic mathematics
 * allowing the exact representation of real number
 * e.g. sin(2*pi/3) will be represented as sqrt(3)/2 
"""
class coordinate():
    """
    setting the precision of the floating point conversion to 100
    i.e. the value of the coordinate is rounded to 
         100 decimal places when converted to float value
    """
    getcontext().prec = 100
    
    def __init__(self, num):
        if isinstance(num, int) or isinstance(num, float):
            a = Fraction(Decimal(str(num)))
            self.value = sympify(a)
            #self.decimal = Decimal(str(num))
        elif isinstance(num, coordinate):
            self.value = num.value
            #self.decimal = num.decimal
        else:
            self.value = num            
    
    def __float__(self):
        #raise NotImplementedError
        return float(self.value)
        
    def __repr__(self):
        return str(self.value)
        
    def __str__(self):
        return str(self.value)
        
    """
    simplifying the expressions of the coordinate's values
    e.g. sin(2*pi/3) will be simplified into sqrt(3)/2
    * this function is rather inefficient in running time, 
    * try to minimize the use of this function 
    """
    def simplify(self):
        return coordinate(simplify(self.value))
        
    """
    expanding the expressions of the coordinate's values
    e.g. 2*(sqrt(3)+ sqrt(5)) --> 2*sqrt(3) + 2*sqrt(5)
    * this function is rather inefficient in running time, 
    * try to minimize the use of this function 
    """
    def expand(self):
        return coordinate(expand(self.value))
        
    """
    converting the value of the coordinate into float value
    """
    def evalf(self):
        return self.value.evalf()
        
    """
    finding the sign of the coordinate: 1 if positive, -1 if negative, 0 if 0
    """
    def sign(self):
        if self.value > 0:
            return coordinate(1)
        elif self.value < 0:
            return coordinate(-1)
        return coordinate(0)
        
    """
    finding the sign of the coordinate: 1 if positive, -1 if negative, 0 if 0
    """
    def copy_sign(self, num=1):
        if self.value > 0:
            return 1
        elif self.value < 0:
            return -1
        return 0
        
    """
    check if the coordinate is equal to 0
    """
    def isZero(self):
        return self.value == 0
    """
    check if the coordinate is equal to 1
    """
    def isOne(self):
        return self.value == 1
    
    """
    finding square root of the coordinate
    """
    def sqrt(self):
        return coordinate(sqrt(self.value))
        
    """
    additioin
    """
    def __add__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value + other.value)   

    """
    subtraction
    """
    def __sub__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value - other.value)

    """
    multiplication
    """
    def __mul__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value * other.value)

    """
    division
    """
    def __truediv__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value / other.value)   

    """
    floor dividsion (only taking the integer part of the result)
    e.g. 23 // 2 = 11
    """
    def __floordiv__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(int(self.value / other.value))  

    """
    power
    """
    def __pow__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value ** other.value) 

    """
    modular
    """
    def __mod__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value % other.value)   

    """
    negation
    """
    def __neg__ (self):
        return coordinate(-self.value)   

    """
    """
    def __pos__ (self):
        return  coordinate(self.value)

    """
    absolute value
    """
    def __abs__(self):
        return coordinate(abs(self.value))

    """
    inverting
    """
    def invert (self):
        return coordinate(S('1') / self.value)   

    """
    check equality
    """
    def __eq__(self, other):
        if isinstance(other, coordinate):
            return self.value.evalf() == other.value.evalf()
        elif isinstance(other, float) or isinstance(other, int):
            return self.value.evalf() == other
        return self.value.evalf() == other.evalf()

    """
    <
    """
    def __lt__(self,other):
        if isinstance(other, coordinate):
            return self.value.evalf() < other.value.evalf()
        elif isinstance(other, float) or isinstance(other, int):
            return self.value.evalf() < other
        return self.value.evalf() < other.evalf()

    """
    >
    """
    def __gt__(self,other):
        if isinstance(other, coordinate):
            return self.value.evalf() > other.value.evalf()
        elif isinstance(other, float) or isinstance(other, int):
            return self.value.evalf() > other
        return self.value.evalf() > other.evalf()

    """
    <=
    """
    def __le__(self,other):
        if isinstance(other, coordinate):
            return self.value.evalf() <= other.value.evalf()
        elif isinstance(other, float) or isinstance(other, int):
            return self.value.evalf() <= other
        return self.value.evalf() <= other.evalf()

    """
    >=
    """
    def __ge__(self,other):
        if isinstance(other, coordinate):
            return self.value.evalf() >= other.value.evalf()
        elif isinstance(other, float) or isinstance(other, int):
            return self.value.evalf() >= other
        return self.value.evalf() >= other.evalf()
        

    """
    comparing float values
    """
    def compareFloat(self, other):
        if not isinstance(other, coordinate):
            if isinstance(other, float) or isinstance(other, int):
                if self.value.evalf() == other:
                    return 0
                elif self.value.evalf() > other:
                    return 1
                return -1
            else:
                if self.value.evalf() == other.evalf():
                    return 0
                elif self.value.evalf() > other.evalf():
                    return 1
                return -1
        if self.value.evalf() == other.value.evalf():
            return 0
        elif self.value.evalf() > other.value.evalf():
            return 1
        return -1

    """
    comparing values
    """
    def compareValue(self,other):
        if isinstance(other, coordinate):
            if self.value.evalf() == other.value.evalf():
                return 0
            elif self.value.evalf() > other.value.evalf():
                return 1
            return -1
        elif isinstance(other, float) or isinstance(other, int):
            if self.value.evalf() == other:
                return 0
            elif self.value.evalf() > other:
                return 1
            return -1
        if self.value.evalf() == other.evalf():
            return 0
        elif self.value.evalf() > other.evalf():
            return 1
        return -1
