# -*- coding: utf-8 -*-
"""
Created on Tue Nov  8 12:39:31 2016

@author: akibdpt
"""

from .vectorFloat import vector2Dfloat

class sSeg():
    def __init__(self, segment, other = None):
        if segment.type != 'struct-bSegment':
            print(segment.type)
            raise ValueError('segment is not a bSegment')
        self.type = 'struct-sSeg'
        if other is not None:
            self.segment = segment
            self.membership = other.membership
            self.alignedWithOne = other.alignedWithOne
            self.alignedWithOther = other.alignedWithOther
        else:
            self.segment = segment
            self.membership = False
            self.alignedWithOne = False
            self.alignedWithOther = False
            
    def __str__(self):
        return 'ss-'+str(self.segment)
            
    def splayCompare(self, other):
        p = vector2Dfloat.p
        if self.segment.start.type == 'struct-vector2D' or self.segment.start.type == 'struct-vector2Dfloat':
            
            if (other.type == 'struct-vector2D' or other.type == 'struct-vector2Dfloat' ):
                return round(float(other.subtract(self.segment.start).product(self.segment.direction()) * -1),p)
            elif not other.type == 'struct-sSeg':
                raise TypeError('other must be either splaySegment or vector')
            c = round(float(other.segment.start.subtract(self.segment.start).product(self.segment.direction()) * -1),p)
            if c != 0:
                return c
            return round(float(other.segment.end.subtract(self.segment.start).product(self.segment.direction()) * -1),p)
        else:
            if other.type == 'struct-vector3D' or other.type == 'struct-vector3Dfloat':
                return round(float(other.subtract(self.segment.start).product(self.segment.direction()).sign() * -1),p)
            elif not other.type == 'struct-sSeg':
                raise TypeError('other must be either splaySegment or vector')
            c = round(float(other.segment.start.subtract(self.segment.start).product(self.segment.direction()).sign() * -1),p)
            if c != 0:
                return c
            return round(float(other.segment.end.subtract(self.segment.start).product(self.segment.direction()).sign() * -1),p)

class sRef():
    START_POSITION = 0
    END_POSITION = 1
    INTERSECTION = 2
    NOT_PART_OF_CYCLE = 0
    PART_OF_CYCLE = 1
    
    def __init__(self, position, segment, tp):
        self.type = 'struct-sRef'
        if 'ind' in position.type:
            raise ValueError('position must not be an individual')
        self.position = position
        self.segment = segment
        self.posType = tp
        
    def __str__(self):
        return 'ref'+str(self.position) +'..'+ str(self.segment)
        
    def isStartPosition(self):
        return self.posType == sRef.START_POSITION
        
    def isEndPosition(self):
        return self.posType == sRef.END_POSITION
        
    def isIntersection(self):
        return self.posType == sRef.INTERSECTION
        
    def isPartOfCycle(self):
        return self.posType == sRef.PART_OF_CYCLE

    @staticmethod
    def startPosition(seg):
        return sRef(seg.segment.start, seg, sRef.START_POSITION)
        
        
    def setPartOfCycle(self, partOfCycle):
        if (partOfCycle):
            self.posType = sRef.PART_OF_CYCLE
        else: self.posType = sRef.NOT_PART_OF_CYCLE

    @staticmethod
    def endPosition(seg):
        return sRef(seg.segment.end, None, sRef.END_POSITION)

    @staticmethod
    def intersection(vect):
        return sRef(vect, None, sRef.INTERSECTION)

    def splayCompare(self, other):
        if 'struct-s' in self.position.type:
            if other.type == 'struct-sRef':
                return self.position.splayCompare(other.position)
            else: 
                return self.position.splayCompare(other)
        if other.type == 'struct-sRef':
            return self.position.compareValue(other.position)
        return self.position.compareValue(other)
        
class vertex(sRef):
    vertices = {}
    
    def __init__(self, vect):
        if (vect.type != 'struct-vector2D' and vect.type != 'struct-vector2Dfloat' and
            vect.type != 'struct-vector3D' and vect.type != 'struct-vector3Dfloat'):
            raise ValueError('vect is not a vector')
        super().__init__(vect, None, 0)
        self.type = 'struct-sVertex'
        self.edges = []
        string = str(vect.x)+'|'+str(vect.y)
        vertex.vertices[string] = self
        #vertex.vertices[vect] = self


    def lastVisited(self):
        return self.segment
        
    def setLastVisited(self, interval):
        self.segment = interval
        
    def __eq__(self, other):
        if other.type != 'struct-sVertex':
            return False
        return self.position == other.position
        
    def compareValue(self, other):
        return 0
